#ifndef __Feature_hpp__
#define __Feature_hpp__

#include "IntegerRange.hpp"
#include <ostream>
#include <vector>

///
/// @brief A tracked feature.
///
class Feature {

public:

  ///
  /// @brief A location of a feature inside a frame.
  ///
  struct Location {
    int projectionId_;
    float x_;
    float y_;

    Location() {};
    Location( int projectionId, float x, float y ): 
		 projectionId_( projectionId ), x_( x ), y_( y ) {}
  };

private:

  int id_;                              ///< The ID of the feature.
  int objectId_;                        ///< The ID of the object.
  size_t firstViewFrame_;               ///< index of 1st frame to view feature.
  std::vector<Location> locations_;     ///< locations of feature in frames.

public:

  ///
  /// @brief Initialize a new instance of the Feature class with an ID of -1.
  ///
  Feature();

  ///
  /// @brief Initialize a new instance of the Feature class with a 
  /// specified ID, first viewing frame index, and set of locations.
  ///
  Feature( int id,size_t firstViewFrame,const std::vector<Location> locations );

  // Delete this when it's no longer needed for compilation.
  Feature( int id, int objectId_, size_t firstViewFrame, 
		  const std::vector<Location> locations );

  ///
  /// @brief Initialize a new instance of the Feature class as a copy of 
  /// another instance.
  ///
  Feature( const Feature & feature );

  ///
  /// @brief Assignment operator.
  ///
  const Feature & operator=( const Feature & feature );

  ///
  /// @brief Get the ID of the feature instance.
  ///
  int GetID() const;

  ///
  /// @brief Get the ID of the feature that is used to reconstruct this feature.
  ///
  int GetPartnerID() const;

  ///
  /// @brief Set the ID of the feature that is used to reconstruct this feature.
  ///
  void SetPartnerID( int partnerId );

  ///
  /// @brief Get the ID of the underlying object.
  ///
  int GetObjectID() const;

  ///
  /// @brief Get the projected location of the feature at a given frame.
  ///
  /// The returned reference is valid only as long as the Feature instance exists.
  ///
  const Location & GetLocation( int frame ) const;



  void overwriteLocationList(int newFirstFrame, std::vector<Location> locList);

  ///
  /// @brief Get the range of frames in which the feature is seen.
  ///
  const IntegerRange GetRange() const;

  ///
  /// @brief Get the number of frames in which the feature is seen.
  ///
  int GetRangeSize() const;

};

///
/// @brief Output a Feature instance to an ostream instance.
///
std::ostream & operator<<( std::ostream & os, const Feature & feature );

#endif // __Feature_hpp__
