#ifndef __IntegerSet_hpp__
#define __IntegerSet_hpp__

#include "IntegerRange.hpp"
#include <set>

class IntegerSet {

  friend std::ostream & operator<<( std::ostream & os, 
		  const IntegerSet & integerSet );

  ///
  /// @brief Compare two IntegerRange instances by performing a '<' operation 
  /// on the lower bounds.
  ///
  struct CompareIntegerRange {
    bool operator()( const IntegerRange & integerRange0, 
			 const IntegerRange & integerRange1 ) const;
  };

  typedef std::set<IntegerRange, CompareIntegerRange> IntegerRangeSet;

  IntegerRangeSet ranges_;              ///< The ranges contained in the set.
  int size_;                            ///< The number of integers in the set.

public:

  ///
  /// @brief Initialize a new instance of the IntegerSet class as the empty set.
  ///
  IntegerSet();

  ///
  /// @brief Initalize a new instance of the IntegerSet class as a copy 
  /// of another instance.
  ///
  IntegerSet( const IntegerSet & integerSet );

  ///
  /// @brief Initalize a new instance of the IntegerSet class to contain 
  /// a specified range of integers.
  ///
  IntegerSet( const IntegerRange & integerRange );

  ///
  /// @brief Assignment operator.
  ///
  const IntegerSet & operator=( const IntegerSet & integerSet );

  ///
  /// @brief Set addition operator (union).
  ///
  const IntegerSet & operator+=( const IntegerSet & integerSet );

  ///
  /// @brief Set subtraction operator (intersection with the complement).
  ///
  const IntegerSet & operator-=( const IntegerSet & integerSet );

  ///
  /// @brief Set intersection operator.
  ///
  const IntegerSet & operator&=( const IntegerSet & integerSet );

  ///
  /// @brief Equality operator.
  ///
  bool operator==( const IntegerSet & integerSet ) const;

  ///
  /// @brief Inequality operator.
  ///
  bool operator!=( const IntegerSet & integerSet ) const;

  ///
  /// @brief Get the size of the set.
  ///
  int GetSize() const;

};

///
/// @brief Binary set addition (union) operator.
///
const IntegerSet operator+( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 );

///
/// @brief Binary set subtraction operator.
///
const IntegerSet operator-( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 );

///
/// @brief Binary set intersection operator.
///
const IntegerSet operator&( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 );

///
/// @brief Output an IntegerSet instance to an ostream instance.
///
std::ostream & operator<<( std::ostream & os, const IntegerSet & integerSet );

#endif // __IntegerSet_hpp__
