#ifndef VM2VECTOR_ALREADY_INCLUDED
#define VM2VECTOR_ALREADY_INCLUDED

#include <common.h>

#define VM2VECTOR_X         0
#define VM2VECTOR_Y         1
#define VM2VECTOR_RAD_TO_DEG(R)        (((R)/M_PI)*180.0)
#define VM2VECTOR_DEG_TO_RAD(D)        (((D)/180.0)*M_PI)

class VM2Point;
class VM2Matrix;
class VM3Matrix;


/*****************************************************************************

The VM2Vector class is used to store two component vectors. Most
of the code is inline to optimize performance, though increasing code size.

*****************************************************************************/

class VM2Vector {

   private:
      float vx, vy;

   public:

      /*
       * constructors
       */
      VM2Vector() 
      { 
         vx = vy = 0.0;
      }

      VM2Vector(VM2Point &);

      VM2Vector(const VM2Vector &ref) 
      {
         vx = ref.vx; vy = ref.vy;
      }

      VM2Vector(int ix, int iy)
      { 
         vx = (float)ix; vy = (float)iy; 
      }

      VM2Vector(float fx, float fy)
      {
         vx = fx; vy = fy; 
      }

      VM2Vector(double fx, double fy)
      {
         vx = fx; vy = fy; 
      }

      VM2Vector(float v[2]) 
      {
         vx = v[0]; vy = v[1];
      }


      /*
       * utilities
       */
      inline float dot(const VM2Vector &v) 
      {
        return (vx*v.vx + vy*v.vy);
      }

      inline void normalize() 
      {
         float vlength = length();
         vx /= vlength; vy /= vlength;
      }

      inline float angleBetween(const VM2Vector &v, int normalized=FALSE)
      {
         float dotProduct;

         if (normalized) {
            dotProduct = this->dot(v);
         } else {
            VM2Vector a,b;
            a = *this;
            b = v;
            a.normalize();
            b.normalize();
            dotProduct = a.dot(b);
         }
         return (ACOS(dotProduct));
      }

      inline float heading()
      {
         float dir[4][2] = { {1,0}, {0,1}, {-1,0}, {0,-1} };
         float len = SQRT(SQR(vx) + SQR(vy));
         float nvx = vx / len;
         float nvy = vy / len;
         float dotProduct, radians;
         int d = -1;

         if      (nvx >= 0 && nvy >= 0) d = 0;
         else if (nvx <= 0 && nvy >= 0) d = 1;
         else if (nvx <= 0 && nvy <= 0) d = 2;
         else if (nvx >= 0 && nvy <= 0) d = 3;
         dotProduct = dir[d][0]*nvx + dir[d][1] * nvy;
         CLAMP(dotProduct,0,1);
         radians = ACOS(dotProduct) + (float)d*M_PI_2;

         return (radians);
      }

      inline void set(float sx, float sy)
      {
         vx = sx; vy = sy;
      }

      inline void negate() 
      {
         vx = -vx; vy = -vy; 
      }

      inline void print(FILE *fp) 
      {
         fprintf(fp,"(%1.4f, %1.4f)\n",vx,vy);
      }

      inline float x() const { return (vx); }
      inline float y() const { return (vy); }

      inline float length() 
      { 
         return (SQRT(vx*vx+vy*vy));
      }

      inline float lengthSquared() 
      { 
         return (vx*vx+vy*vy);
      }

      inline VM2Vector perpendicular()
      {
         return (VM2Vector(-vy, vx));
      }

      inline int isZero() 
      { 
         return ((ABS(vx - 0) < FLOAT_EPSILON) &&
                 (ABS(vy - 0) < FLOAT_EPSILON));
      }


      /*
       * operator overloading
       */

      inline float operator()(int r) const
      { 
         float val;
         switch (r) {
            case VM2VECTOR_X : val = vx; break;
            case VM2VECTOR_Y : val = vy; break;
            default: fprintf(stderr, 
                             "VM2Vector: coordinate index out of range\n");
                     val = 0;
         }
         return (val);
      }

      inline VM2Vector &operator=(const VM2Vector& v) 
      {
         vx = v.vx; vy = v.vy;
         return *this;
      }

      inline VM2Vector &operator+=(const VM2Vector& a) 
      {
         vx += a.vx; vy += a.vy;
         return *this;
      }

      inline VM2Vector &operator-=(const VM2Vector& a) 
      {
         vx -= a.vx; vy -= a.vy;
         return *this;
      }

      inline friend VM2Vector operator+(const VM2Vector& a, const VM2Vector& b) 
      {
         return VM2Vector(a.vx+b.vx, a.vy+b.vy);
      }

      inline friend VM2Vector operator-(const VM2Vector& a, const VM2Vector& b) 
      {
         return VM2Vector(a.vx-b.vx, a.vy-b.vy);
      } 

      inline friend VM2Vector operator*(const VM2Vector& a, int s) 
      {
         return VM2Vector(a.vx*(float)s, a.vy*(float)s);
      }

      inline friend VM2Vector operator*(const VM2Vector& a, float s) 
      {
         return VM2Vector(a.vx*s, a.vy*s);
      };

      inline friend VM2Vector operator*(const VM2Vector& a, double s) 
      {
         return VM2Vector(a.vx*s, a.vy*s);
      }

      inline friend VM2Vector operator*(int s, const VM2Vector& a) 
      {
         return VM2Vector(a.vx*(float)s, a.vy*(float)s);
      }

      inline friend VM2Vector operator*(float s, const VM2Vector& a) 
      {
         return VM2Vector(a.vx*s, a.vy*s);
      }

      inline friend VM2Vector operator*(double s, const VM2Vector& a) 
      {
         return VM2Vector(a.vx*s, a.vy*s);
      }

      inline friend VM2Vector operator/(const VM2Vector& a, int s) 
      {
         return VM2Vector(a.vx/(float)s, a.vy/(float)s);
      }

      inline friend VM2Vector operator/(const VM2Vector& a, float s) 
      {
         return VM2Vector(a.vx/s, a.vy/s);
      }

      inline friend VM2Vector operator/(const VM2Vector& a, double s) 
      {
         return VM2Vector(a.vx/s, a.vy/s);
      }

      /***********/

      friend VM2Vector operator*(const VM2Matrix& m, const VM2Vector &v);
      friend VM2Vector operator*(const VM3Matrix& m, const VM2Vector &v);

      /***********/

      inline int operator==(const VM2Vector& v) {
         return (ABS(vx - v.vx) < FLOAT_EPSILON && 
                 ABS(vy - v.vy) < FLOAT_EPSILON);
      }

      inline int operator!=(const VM2Vector& v) {
         return (ABS(vx - v.vx) > FLOAT_EPSILON ||
                 ABS(vy - v.vy) > FLOAT_EPSILON);
      }

      /*
       * Destructor
       */
      ~VM2Vector() {};
};

#endif
