#ifndef VM3MATRIX_ALREADY_INCLUDED
#define VM3MATRIX_ALREADY_INCLUDED

#include <vm2vector.h>
#include <vm2point.h>


#define VM3MATRIX_ROTATE_X_AXIS        0
#define VM3MATRIX_ROTATE_Y_AXIS        1
#define VM3MATRIX_ROTATE_Z_AXIS        2

#define VM3MATRIX_RAD_TO_DEG(R)        (((R)/M_PI)*180.0)
#define VM3MATRIX_DEG_TO_RAD(D)        (((D)/180.0)*M_PI)


/*****************************************************************************

This class implements a standard 3x3 matrix. Operations are assumed to be 
performed with homogenous 2D coordinates (using VM2Vector/VM2Point). Supported
operations are: translate, scale, rotate, invert.

*****************************************************************************/

class VM3Matrix {
   private:
      float mat[3][3];

   public:

      /* 
       * Constructors 
       */
      VM3Matrix() 
      { 
         mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0;
         mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0;
      }

      VM3Matrix(float a, float b, float c, 
                float d, float e, float f,
                float g, float h, float i)
      {
         mat[0][0] = a; mat[0][1] = b; mat[0][2] = c;
         mat[1][0] = d; mat[1][1] = e; mat[1][2] = f;
         mat[2][0] = g; mat[2][1] = h; mat[2][2] = i;
      }

      VM3Matrix(float m[3][3]) 
      {
         BCOPY(m, mat, sizeof(float)*9);
      }

      VM3Matrix(const VM3Matrix &m) 
      { 
         BCOPY(m.mat, mat, sizeof(float)*9);
      }


      /*
       * Matrix operations and utilities
       */
      void identity() 
      {
         mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0;
         mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0;
      }

      void zero() 
      {
         BZERO(mat, sizeof(float)*9);
      }

      void transpose() 
      {
        int   r,c;
        float tmp;

        for (r = 0; r < 3; r++) {
           for (c = r; c < 3; c++) {
              if (r != c) {
                 SWAP(mat[r][c], mat[c][r], tmp);
              }
           }
        }
      }

      /* translation operation */
      void translate(float x, float y)
      {
         mat[0][2] += x;
         mat[1][2] += y;
      }
      void translate(VM2Point  &t) { translate(t.x(),t.y()); }
      void translate(VM2Vector &t) { translate(t.x(),t.y()); }

      /* scale operation */
      void scale(float x, float y)
      {
         VM3Matrix mat;

         mat.mat[0][0] = x;
         mat.mat[1][1] = y;

         *this = (*this) * mat;
      }
      void scale(const VM2Point  &s) { scale(s.x(),s.y()); }
      void scale(const VM2Vector &s) { scale(s.x(),s.y()); }
      void scale(float s)      { scale(s, s); }

      /* rotation operation */
      void rotate(float radians);

      /* inversion operation */
      int  invert();

      /* others */
      void set(int r, int c, float val) 
      { 
         IFDEBUG(r < 0 || r >= 3 || c < 0 || c >= 3) {
            fprintf(stderr, "VM3Matrix: index (%i,%i) out of range\n", r, c);
            return;
         }
         mat[r][c] = val; 
      }

      void print(FILE *fp)
      {
         int i,j;

         for (i=0; i<3; i++) {
            for (j=0; j<3; j++)
               fprintf(fp,"%1.4f ", mat[i][j]);
            fprintf(fp,"\n");
         }
         fprintf(fp,"\n");
      }



      /*
       * Operator overloading
       */
      inline float operator()(int r, int c) const
      {
         IFDEBUG(r < 0 || r >= 3 || c < 0 || c >= 3) {
            fprintf(stderr, "VM3Matrix: index (%i,%i) out of range\n", r, c);
            return 0;
         }
         return (mat[r][c]);
      }

      inline VM3Matrix &operator=(const VM3Matrix &m) 
      {
         BCOPY(m.mat, mat, sizeof(float)*9);
         return *this;
      }

      inline VM3Matrix &operator+=(const VM3Matrix &m) 
      {
         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             mat[i][j] += m.mat[i][j];

         return *this;
      }

      inline VM3Matrix &operator-=(const VM3Matrix &m) 
      {
         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             mat[i][j] -= m.mat[i][j];

         return *this;
      }

      inline VM3Matrix &operator*=(const VM3Matrix &m) 
      {
         register float res[3][3];

         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             res[i][j] = mat[i][0] * m.mat[0][j] +
                         mat[i][1] * m.mat[1][j] +
                         mat[i][2] * m.mat[2][j];
         BCOPY(res, mat, sizeof(float)*9);
         return *this;
      }

      inline friend VM3Matrix operator+(const VM3Matrix &a, const VM3Matrix &b) 
      {
         register float res[3][3];

         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             res[i][j] = a.mat[i][j] + b.mat[i][j];
         return (VM3Matrix(res));
      }

      inline friend VM3Matrix operator-(const VM3Matrix &a, const VM3Matrix &b) 
      {
         register float res[3][3];

         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             res[i][j] = a.mat[i][j] - b.mat[i][j];
         return (VM3Matrix(res));
      }

      inline friend VM3Matrix operator*(const VM3Matrix &a, const VM3Matrix &b)
      {
         register float res[3][3];

         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             res[i][j] = a.mat[i][0] * b.mat[0][j] +
                         a.mat[i][1] * b.mat[1][j] +
                         a.mat[i][2] * b.mat[2][j];
         return (VM3Matrix(res));
      }

      inline friend VM3Matrix operator*(const VM3Matrix &a, float b)
      {
         register float res[3][3];

         for (int i=0; i<3; i++)
          for (int j=0; j<3; j++)
             res[i][j] = a.mat[i][j] * b;
         return VM3Matrix(res);
      }

      inline int operator==(const VM3Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; ret && i<3; i++)
            for (int j=0; ret && j<3; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = FALSE;
         return (ret);
      }

      inline int operator!=(const VM3Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; !ret && i<3; i++)
            for (int j=0; !ret && j<3; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = TRUE;
         return (ret);
      }


      /*
       * Destructors
       */
      ~VM3Matrix() {};
};


#endif
