#ifndef VM3POINT_ALREADY_INCLUDED
#define VM3POINT_ALREADY_INCLUDED

#include <common.h>

#define VM3POINT_X       0
#define VM3POINT_Y       1
#define VM3POINT_Z       2

class VM3Vector;
class VM3Matrix;
class VM4Matrix;


/*****************************************************************************

This object stores the information that defines a point in 3 space.

*****************************************************************************/

class VM3Point {
   private:
      float px, py, pz;

   public:
      /*
       * constructors are overloaded
       */
      VM3Point() 
      {
         px = py = pz = 0.0;
      }

      VM3Point(VM3Vector &);

      VM3Point(const VM3Point &ref) 
      { 
         px = ref.px; py = ref.py; pz = ref.pz;
      }

      VM3Point(int x, int y, int z) 
      { 
         px = (float)x; py = (float)y; pz = (float)z;
      }

      VM3Point(float x, float y, float z) 
      { 
         px = x; py = y; pz = z;
      }

      VM3Point(double x, double y, double z) 
      { 
         px = x; py = y; pz = z;
      }

      VM3Point(float p[3]) 
      {
         px = p[0]; py = p[1]; pz = p[2];
      }


      /*
       * utilities
       */ 
      inline void set(float sx, float sy, float sz) 
      {
         px = sx; py = sy; pz = sz;
      }

      inline void negate() 
      {
         px = -px; py = -py; pz = -pz;
      }

      inline void print(FILE *fp) 
      {
         fprintf(fp,"(%1.4f, %1.4f, %1.4f)\n",px,py,pz);
      }

      inline float x() const { return px; }; 
      inline float y() const { return py; }; 
      inline float z() const { return pz; }; 

      inline int isZero() 
      { 
         return ((ABS(px - 0) < FLOAT_EPSILON) &&
                 (ABS(py - 0) < FLOAT_EPSILON) &&
                 (ABS(pz - 0) < FLOAT_EPSILON));
      }


      /*
       * operator overloading
       */
      inline float operator()(int r) const
      { 
         float val;
         switch (r) {
            case VM3POINT_X : val = px; break;
            case VM3POINT_Y : val = py; break;
            case VM3POINT_Z : val = pz; break;
            default: fprintf(stderr, 
                             "VM3Point: coordinate index out of range\n");
                     val = 0;
         }
         return (val);
      }

      inline VM3Point &operator=(const VM3Point &c) 
      {
         px = c.px; py = c.py; pz = c.pz;
         return *this;
      }

      inline VM3Point &operator+=(const VM3Point& a) 
      {
         px += a.px; py += a.py; pz += a.pz;
         return *this;
      }

      inline VM3Point &operator-=(const VM3Point& a) 
      {
         px -= a.px; py -= a.py; pz -= a.pz;
         return *this;
      }

      inline friend VM3Point operator+(const VM3Point& a, const VM3Point& b) 
      {
         return VM3Point(a.px+b.px, a.py+b.py, a.pz+b.pz);
      }

      inline friend VM3Point operator-(const VM3Point& a, const VM3Point& b) 
      {
         return VM3Point(a.px-b.px, a.py-b.py, a.pz-b.pz);
      } 

      inline friend VM3Point operator*(const VM3Point& a, int s) 
      {
         return VM3Point(a.px*(float)s, a.py*(float)s, a.pz*(float)s); 
      }

      inline friend VM3Point operator*(const VM3Point& a, float s) 
      {
         return VM3Point(a.px*s, a.py*s, a.pz*s); 
      }

      inline friend VM3Point operator*(const VM3Point& a, double s) 
      {
         return VM3Point((float)(a.px*s), (float)(a.py*s), (float)(a.pz*s)); 
      }

      inline friend VM3Point operator*(int s, const VM3Point& a) 
      {
         return VM3Point(a.px*(float)s, a.py*(float)s, a.pz*(float)s); 
      }

      inline friend VM3Point operator*(float s, const VM3Point& a) 
      {
         return VM3Point(a.px*s, a.py*s, a.pz*s); 
      }

      inline friend VM3Point operator*(double s, const VM3Point& a) 
      {
         return VM3Point((float)(a.px*s), (float)(a.py*s), (float)(a.pz*s)); 
      }

      inline friend VM3Point operator/(const VM3Point& a, int s) 
      {
         return VM3Point(a.px/(float)s, a.py/(float)s, a.pz/(float)s);
      }

      inline friend VM3Point operator/(const VM3Point& a, float s) 
      {
         return VM3Point(a.px/s, a.py/s, a.pz/s);
      }

      inline friend VM3Point operator/(const VM3Point& a, double s) 
      {
         return VM3Point(a.px/s, a.py/s, a.pz/s);
      }


      /************/

      friend VM3Point operator*(const VM4Matrix& m, const VM3Point &p);
      friend VM3Point operator*(const VM3Matrix& m, const VM3Point &p);

      /************/

      inline int operator==(const VM3Point& p) {
         return (ABS(px - p.px) < FLOAT_EPSILON && 
                 ABS(py - p.py) < FLOAT_EPSILON &&
                 ABS(pz - p.pz) < FLOAT_EPSILON);
      }

      inline int operator!=(const VM3Point& p) {
         return (ABS(px - p.px) > FLOAT_EPSILON ||
                 ABS(py - p.py) > FLOAT_EPSILON ||
                 ABS(pz - p.pz) > FLOAT_EPSILON);
      }

      inline int operator>=(const VM3Point& v) 
      {
         return ((px >= v.px) && (py >= v.py) && (pz >= v.pz)) ? TRUE : FALSE;
      }

      inline int operator<=(const VM3Point& v) 
      {
         return ((px <= v.px) && (py <= v.py) && (pz <= v.pz)) ? TRUE : FALSE;
      }

      inline int operator<(const VM3Point& v) 
      {
         return ((px < v.px) && (py < v.py) && (pz < v.pz)) ? TRUE : FALSE;
      }

      inline int operator>(const VM3Point& v) 
      {
         return ((px > v.px) && (py > v.py) && (pz > v.pz)) ? TRUE : FALSE;
      }
};

#endif
