#ifndef VM3VECTOR_ALREADY_INCLUDED
#define VM3VECTOR_ALREADY_INCLUDED

#include <common.h>

#define VM3VECTOR_X         0
#define VM3VECTOR_Y         1
#define VM3VECTOR_Z         2
#define VM3VECTOR_RAD_TO_DEG(R)        (((R)/M_PI)*180.0)
#define VM3VECTOR_DEG_TO_RAD(D)        (((D)/180.0)*M_PI)

class VM3Point;
class VM4Matrix;


/*****************************************************************************

The VM3Vector class is used to store three component vectors. Most
of the code is inline to optimize performance, though increasing code size.

*****************************************************************************/

class VM3Vector {

   private:
      float vx, vy, vz;

   public:

      /*
       * constructors
       */
      VM3Vector() 
      { 
         vx = vy = vz = 0.0;
      }

      VM3Vector(VM3Point &);

      VM3Vector(const VM3Vector &ref) 
      {
         vx = ref.vx; vy = ref.vy; vz = ref.vz;
      }

      VM3Vector(int ix, int iy, int iz) 
      { 
         vx = (float)ix; vy = (float)iy; vz = (float)iz;
      }

      VM3Vector(float fx, float fy, float fz) 
      {
         vx = fx; vy = fy; vz = fz;
      }

      VM3Vector(double fx, double fy, double fz) 
      {
         vx = fx; vy = fy; vz = fz;
      }

      VM3Vector(float v[3]) 
      {
         vx = v[0]; vy = v[1]; vz = v[2];
      }


      /*
       * utilities
       */
      inline VM3Vector cross(const VM3Vector &v) 
      {
        return VM3Vector(vy*v.vz - vz*v.vy, 
                         vz*v.vx - vx*v.vz, 
                         vx*v.vy - vy*v.vx);
      }

      inline float dot(const VM3Vector &v) 
      {
        return (vx*v.vx + vy*v.vy + vz*v.vz);
      }

      inline void normalize() 
      {
         float vlength = length();
         vx /= vlength; vy /= vlength; vz /= vlength;
      }

      inline float angleBetween(const VM3Vector &v, int normalized=FALSE)
      {
         float dotProduct;

         if (normalized) {
            dotProduct = this->dot(v);
         } else {
            VM3Vector a,b;
            a = *this;
            b = v;
            a.normalize();
            b.normalize();
            dotProduct = a.dot(b);
         }
         return (ACOS(dotProduct));
      }

      inline void set(float sx, float sy, float sz) 
      {
         vx = sx; vy = sy; vz = sz;
      }

      inline void negate() 
      {
         vx = -vx; vy = -vy; vz = -vz;
      }

      inline void print(FILE *fp) 
      {
         fprintf(fp,"(%1.4f, %1.4f, %1.4f)\n",vx,vy,vz);
      }

      inline float x() const { return (vx); }
      inline float y() const { return (vy); }
      inline float z() const { return (vz); }

      inline float length() 
      { 
         return (SQRT(vx*vx+vy*vy+vz*vz));
      }

      inline float lengthSquared() 
      { 
         return (vx*vx+vy*vy+vz*vz);
      }

      VM3Vector perpendicular();

      inline int isZero() 
      { 
         return ((ABS(vx - 0) < FLOAT_EPSILON) &&
                 (ABS(vy - 0) < FLOAT_EPSILON) &&
                 (ABS(vz - 0) < FLOAT_EPSILON));
      }


      /*
       * operator overloading
       */

      inline float operator()(int r) const
      { 
         float val;
         switch (r) {
            case VM3VECTOR_X : val = vx; break;
            case VM3VECTOR_Y : val = vy; break;
            case VM3VECTOR_Z : val = vz; break;
            default: fprintf(stderr, 
                             "VM3Vector: coordinate index out of range\n");
                     val = 0;
         }
         return (val);
      }

      inline VM3Vector &operator=(const VM3Vector& v) 
      {
         vx = v.vx; vy = v.vy; vz = v.vz;
         return *this;
      }

      inline VM3Vector &operator+=(const VM3Vector& a) 
      {
         vx += a.vx; vy += a.vy; vz += a.vz;
         return *this;
      }

      inline VM3Vector &operator-=(const VM3Vector& a) 
      {
         vx -= a.vx; vy -= a.vy; vz -= a.vz;
         return *this;
      }

      inline friend VM3Vector operator+(const VM3Vector& a, const VM3Vector& b) 
      {
         return VM3Vector(a.vx+b.vx, a.vy+b.vy, a.vz+b.vz);
      }

      inline friend VM3Vector operator-(const VM3Vector& a, const VM3Vector& b) 
      {
         return VM3Vector(a.vx-b.vx, a.vy-b.vy, a.vz-b.vz);
      } 

      inline friend VM3Vector operator*(const VM3Vector& a, int s) 
      {
         return VM3Vector(a.vx*(float)s, a.vy*(float)s, a.vz*(float)s); 
      }

      inline friend VM3Vector operator*(const VM3Vector& a, float s) 
      {
         return VM3Vector(a.vx*s, a.vy*s, a.vz*s); 
      };

      inline friend VM3Vector operator*(const VM3Vector& a, double s) 
      {
         return VM3Vector(a.vx*s, a.vy*s, a.vz*s); 
      }

      inline friend VM3Vector operator*(int s, const VM3Vector& a) 
      {
         return VM3Vector(a.vx*(float)s, a.vy*(float)s, a.vz*(float)s); 
      }

      inline friend VM3Vector operator*(float s, const VM3Vector& a) 
      {
         return VM3Vector(a.vx*s, a.vy*s, a.vz*s); 
      }

      inline friend VM3Vector operator*(double s, const VM3Vector& a) 
      {
         return VM3Vector(a.vx*s, a.vy*s, a.vz*s); 
      }

      inline friend VM3Vector operator/(const VM3Vector& a, int s) 
      {
         return VM3Vector(a.vx/(float)s, a.vy/(float)s, a.vz/(float)s);
      }

      inline friend VM3Vector operator/(const VM3Vector& a, float s) 
      {
         return VM3Vector(a.vx/s, a.vy/s, a.vz/s);
      }

      inline friend VM3Vector operator/(const VM3Vector& a, double s) 
      {
         return VM3Vector(a.vx/s, a.vy/s, a.vz/s);
      }

      /***********/

      friend VM3Vector operator*(const VM4Matrix& m, const VM3Vector& v);

      /***********/

      inline int operator==(const VM3Vector& v) {
         return (ABS(vx - v.vx) < FLOAT_EPSILON && 
                 ABS(vy - v.vy) < FLOAT_EPSILON &&
                 ABS(vz - v.vz) < FLOAT_EPSILON);
      }

      inline int operator!=(const VM3Vector& v) {
         return (ABS(vx - v.vx) > FLOAT_EPSILON ||
                 ABS(vy - v.vy) > FLOAT_EPSILON ||
                 ABS(vz - v.vz) > FLOAT_EPSILON);
      }

      /*
       * Destructor
       */
      ~VM3Vector() {};
};

#endif
