#include "Disable4786.hpp"
#include "AnchorPlanner.hpp"
#include <algorithm>
#include <cassert>
#include <iostream>

using namespace std;


/*****************************************************************************
*****************************************************************************/
namespace {

  static bool FeatureComparison( const Feature * feature1, 
		                           const Feature * feature2 )
  {
    return feature1->GetRange().GetLowerBound() < 
		     feature2->GetRange().GetLowerBound();
  }

  struct FeaturePairComparison {

    int frame_;
	 float featurePairLifeWeight_, featurePairDistanceWeight_;

    FeaturePairComparison( int frame, float featurePairLifeWeight, 
			 float featurePairDistanceWeight ): 
		 frame_( frame ), featurePairLifeWeight_( featurePairLifeWeight ),
       featurePairDistanceWeight_( featurePairDistanceWeight )
    {}

    bool operator()( const FeaturePair * pair1, const FeaturePair * pair2 ) {
      const Feature::Location & loc11 = 
			pair1->GetFirstFeature().GetLocation( frame_ );
      const Feature::Location & loc12 = 
			pair1->GetSecondFeature().GetLocation( frame_ );
      const Feature::Location & loc21 = 
			pair2->GetFirstFeature().GetLocation( frame_ );
      const Feature::Location & loc22 = 
			pair2->GetSecondFeature().GetLocation( frame_ );

      float dX = loc11.x_ - loc12.x_;
      float dY = loc11.y_ - loc12.y_;
      float WeightedDist1 = featurePairDistanceWeight_ * 
			log( dX * dX + dY * dY ) / 2;
      dX = loc21.x_ - loc22.x_;
      dY = loc21.y_ - loc22.y_;
      float WeightedDist2 = featurePairDistanceWeight_ * 
			log( dX * dX + dY * dY ) / 2;

	  float weightedLife1 = featurePairLifeWeight_ * 
		  log( pair1->GetImageRange().GetSize() );
	  float weightedLife2 = featurePairLifeWeight_ * 
		  log( pair2->GetImageRange().GetSize() );

      return ( WeightedDist1 + weightedLife1 ) > 
			    ( WeightedDist2 + weightedLife2 );
    }

  };

}

/*****************************************************************************
*****************************************************************************/
AnchorPlanner::AnchorPlanner( SceneData & sceneData, 
		const AnchorReconstructor & anchorReconstructor ):
         anchorReconstructor_( anchorReconstructor ), sceneData_( sceneData )
{
  //
  // Create the sorted features list and find the max image ID.
  //
  const std::vector<Feature> & featureList = sceneData_.GetFeatureList();
  sortedFeatureList_.resize( featureList.size() );
  maxImageID_ = 0;
  for( int i=0; i<sortedFeatureList_.size(); ++i ) {
    sortedFeatureList_[i] = &featureList[i];

    if( featureList[i].GetRange().GetUpperBound() > maxImageID_ )
      maxImageID_ = featureList[i].GetRange().GetUpperBound();
  }

  //
  // Sort the features.
  //
  std::sort( sortedFeatureList_.begin(), sortedFeatureList_.end(), 
		  FeatureComparison );

  //
  // Initialize the data structures.
  //
  BuildFeaturePairs();
}


/*****************************************************************************
*****************************************************************************/
void AnchorPlanner::PlanAnchors( float featurePairLifeWeight, 
		float featurePairDistanceWeight, float continuationThreshold, 
		int anchorImageNumber )
{
  const std::vector<bool> & viewpointValidList = 
	  sceneData_.GetViewpointValidOrientationList();

  //
  // Clear the anchor list.
  //
  {
    deque<FeaturePair*> tmp( maxImageID_, 0 );
    anchorList_.swap( tmp );
  }

  //
  // Initialize the previous image ID & pair values.
  //
  int prevImageID = -2;
  FeaturePair * prevAnchorPair = 0;

  //
  // Go through the images and pick the best anchor pair for each.
  //
  for( int imageID=0; imageID < imageFeatureList_.size(); ++imageID ) {

	  //if ((imageID % 100) == 0)
		  //printf("imageID=%i\n",imageID);

    deque<FeaturePair*> & featureList = imageFeatureList_[imageID];

    if( !featureList.empty() ) {
      if( imageID == prevImageID+1 && prevAnchorPair != 0 ) {
        assert( prevAnchorPair != 0 );

        //
        // Check if the same pair is good.  If so, use it and move on.
        //

		float x[2], y[2];

		x[0] = prevAnchorPair->GetFirstFeature().GetLocation( prevImageID ).x_;
		y[0] = prevAnchorPair->GetFirstFeature().GetLocation( prevImageID ).y_;
		x[1] = prevAnchorPair->GetSecondFeature().GetLocation( prevImageID ).x_;
		y[1] = prevAnchorPair->GetSecondFeature().GetLocation( prevImageID ).y_;

		float sqContinuationThreshold = 
			continuationThreshold * continuationThreshold * 
			( ( x[0] - x[1] ) * ( x[0] - x[1] ) + 
			  ( y[0] - y[1] ) * ( y[0] - y[1] ) );

        if( CheckFeaturePair( imageID, *prevAnchorPair, 
					  sqContinuationThreshold ) ) {
          prevImageID = imageID;
          anchorList_[imageID] = prevAnchorPair;
          continue;
        }
      }
    }

    //
    // Create the functor for comparing feature pairs.
    //
    FeaturePairComparison featurePairComparison( imageID, 
			 featurePairLifeWeight, featurePairDistanceWeight );

    //
    // Find the best feature pair that is reconstructable.
    //
    FeaturePair * bestPair = 0;
    if( viewpointValidList[imageID] ) {
      sort( featureList.begin(), featureList.end(), featurePairComparison );
      for( int i=0; i<featureList.size(); ++i ) {
        if( anchorReconstructor_.ReconstructAnchorPair( *featureList[i], 
					  anchorImageNumber ) ) {
          bestPair = featureList[i];
          break;
        }
      }
    }

    if( bestPair != 0 ) {
      anchorList_[imageID] = bestPair;
      prevAnchorPair = bestPair;
      prevImageID = imageID;
    }
    else
      cerr << "Unable to find anchor pair for image " << imageID << endl;

    //while( 1 ) {
      //deque<FeaturePair*>::iterator bestPairIter;
      //bestPairIter = min_element( featureList.begin(), featureList.end(), featurePairComparison );
      //if( bestPairIter == featureList.end() ) {
        //cerr << "Unable to find anchor pair for image " << imageID << endl;
        //break;
      //}

      //FeaturePair * bestPair = *bestPairIter;
      //bestPair->ReconstructPair( sceneData_ );
      //if( bestPair->IsReconstructed() ) {
        //anchorList_[imageID] = bestPair;
        //prevAnchorPair = bestPair;
        //prevImageID = imageID;
        //break;
      //}
      //else {
        //featureList.erase( bestPairIter );
        //cout << "Failed to reconstruct an anchor pair for image " << imageID
          //<< " with " << featureList.size() << " pairs left." << endl;
      //}
    //}
    //cout << "DONE WITH IMAGE " << imageID << "!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
  }
}


/*****************************************************************************
*****************************************************************************/
deque<FeaturePair*> & AnchorPlanner::GetAnchorList()
{
  return anchorList_;
}


/*****************************************************************************
*****************************************************************************/
void AnchorPlanner::BuildFeaturePairs()
{
  //
  // Clear the list of feature pairs.
  //
  if( !featurePairList_.empty() ) {
    deque<FeaturePair> tmp;
    featurePairList_.swap( tmp );
  }

  //
  // Resize the list of each image's feature pairs.
  //
  {
	std::vector<deque<FeaturePair*> > tmp( maxImageID_, deque<FeaturePair*>() );
    imageFeatureList_.swap( tmp );
  }

  //
  // Build the list of feature pairs.
  //
  for( int i=0; i<sortedFeatureList_.size(); ++i ) {
    int currentUpperBound = sortedFeatureList_[i]->GetRange().GetUpperBound();
    for( int j=i+1; j<sortedFeatureList_.size() && 
			 sortedFeatureList_[j]->GetRange().GetLowerBound() < currentUpperBound;
			 ++j ) {
      FeaturePair pair( *sortedFeatureList_[i], *sortedFeatureList_[j] );
      const IntegerRange & imageRange = pair.GetImageRange();
      if( imageRange.GetSize() >= 6 ) {

        //
        // Add the pair to the list of all pairs.
        //
        featurePairList_.push_back( pair );

	    //
        // Add the pair to all the images that 'see' it.
        //
        for(int k=imageRange.GetLowerBound(); k<imageRange.GetUpperBound(); ++k)
          imageFeatureList_[k].push_back( &featurePairList_.back() );
      }
    }
  }

  cout << "Constructed " << featurePairList_.size() << " pairs." << endl;
}


/*****************************************************************************
*****************************************************************************/
bool AnchorPlanner::CheckFeaturePair( int imageID, 
		const FeaturePair & featurePair, float sqContinuationThreshold )
{
  try {
    const Feature::Location & loc1 = 
		 featurePair.GetFirstFeature().GetLocation( imageID );
    const Feature::Location & loc2 = 
		 featurePair.GetSecondFeature().GetLocation( imageID );

    float dX = loc2.x_ - loc1.x_;
    float dY = loc2.y_ - loc1.y_;

    float sqDist = dX * dX + dY * dY;
    return sqDist > sqContinuationThreshold;
  }
  catch( const runtime_error & e ) {
    return false;
  }
}
