#ifndef __AnchorPlanner_hpp__
#define __AnchorPlanner_hpp__

#include "AnchorReconstructor.hpp"
#include "Feature.hpp"
#include "FeaturePair.hpp"
#include "IntegerSet.hpp"
#include "SceneData.hpp"
#include <algorithm>
#include <list>
#include <map>
#include <set>
#include <vector>
#include <deque>

class SceneData;
class AnchorReconstructor;

class AnchorPlanner {

  ///< The scene data.
  SceneData & sceneData_;                                   
  ///< The anchor reconstructor.
  const AnchorReconstructor & anchorReconstructor_;         
  ///< The feature list.
  std::vector<const Feature*> sortedFeatureList_;           
  ///< The pairs of features.
  std::deque<FeaturePair> featurePairList_;                 
  ///< The feature pairs of each image.
  std::vector<std::deque<FeaturePair*> > imageFeatureList_; 
  ///< The anchors chosen for each image.
  std::deque<FeaturePair*> anchorList_;                     
  ///< The largest image id.
  size_t maxImageID_;                                       

  ///
  /// @brief Disallow copying of AnchorPlanner instances.
  ///
  AnchorPlanner( const AnchorPlanner & );

public:

  ///
  /// @brief Initialize a new instance of the AnchorPlanner class with 
  /// a SceneData instance.
  ///
  AnchorPlanner( SceneData & sceneData, 
		  const AnchorReconstructor & anchorReconstructor );

  ///
  /// @brief Plan the features to use as anchors.
  ///
  void PlanAnchors( float featurePairLifeWeight, 
		  float featurePairDistanceWeight, float continuationThreshold, 
		  int anchorImageNumber );

  ///
  /// @brief Get the array of anchor pairs.
  ///
  std::deque<FeaturePair*> & GetAnchorList();

private:

  ///
  /// @brief Build the list of feature pairs.
  ///
  void BuildFeaturePairs();

  ///
  /// @brief Check if a given feature pair is valid for use as anchors for
  /// a given image.
  ///
  static bool CheckFeaturePair( int imageID, const FeaturePair & featurePair, 
		  float sqContinuationThreshold );

};

#endif // __AnchorPlanner_hpp__
