#include <math.h>
#include <stdio.h> 
#include "Nrutil.h"
#include "SceneData.hpp"
#include "FeaturePair.hpp"

#define TOL 2.0e-7
#define ITMAX 400
#define EPS 1.0e-10 
#define FREEALL free_vector(xi,1,n);free_vector(h,1,n);free_vector(g,1,n);
#define GOLD 1.618034 
#define GLIMIT 100.0 
#define TINY 1.0e-20 
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d); 
#define CGOLD 0.3819660 
#define ZEPS 1.0e-10 

static int ncom;
static float *pcom, *xicom, (*nrfunc)( SceneData &, const FeaturePair &, int [], int, float []); 

/*Here ITMAX is the maximum allowed number of iterations, while EPS is a small number to rectify the special case of converging to 
exactly zero function value. */

bool frprmn( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], 
   int imageNumber, float p[], int n, float ftol, int *iter, 
   float *fret, 
   float (*func)(SceneData &,const FeaturePair &,int [],int,float []),
   void (*dfunc)(SceneData &,const FeaturePair &,int [],int,float [],float []))

/*Given a starting point p[1..n], Fletcher-Reeves-Polak-Ribiere minimization is performed on a function func, using its gradient as
calculated by a routine dfunc.  The convergence tolerance on the function value is input as ftol.  Returned quantities are p (the
location of the minimum), iter (the number of iterations that were performed), and fret (the minimum value of the function).  
The routine linmin is called to perform line minimizations.  */

{ 
  void linmin( SceneData & sceneData, const FeaturePair & pair, int imageIDs[],
		  int imageNumber, float p[], float xi[], int n, float *fret, 
        float (*func)( SceneData &, const FeaturePair &, int [], int,float []));
  int j, its; 
  float gg, gam, fp, dgg; 
  float *g, *h, *xi;

  g = vector( 1, n ); 
  h = vector( 1, n ); 
  xi = vector( 1, n ); 

  /*Initializations.  */
  fp = (*func)( sceneData, pair, imageIDs, imageNumber, p ); 
  (*dfunc)( sceneData, pair, imageIDs, imageNumber, p, xi ); 

  for ( j=1; j<=n; j++ ) 
  { g[j] = -xi[j];
    xi[j] = h[j] = g[j]; 
  } 

  /*Loop over iterations.*/
  for ( its=1; its<=ITMAX; its++ ) 
  { *iter = its; 
    linmin( sceneData, pair, imageIDs, imageNumber, p, xi, n, fret, func );

	if ( 2.0 * fabs( *fret - fp ) <= ftol * ( fabs( *fret ) + fabs( fp ) + EPS ) ) 
    { FREEALL
      return true; 
    } 

    fp = *fret; 
    (*dfunc)( sceneData, pair, imageIDs, imageNumber, p, xi );
    dgg = gg = 0.0; 

    for ( j=1; j<=n; j++ ) 
    { gg += g[j] * g[j];
      /* dgg += xi[j] * xi[j];  This statement for Fletcher-Reeves. */
      dgg += ( xi[j] + g[j] ) * xi[j]; /*This statement for Polak-Ribiere. */
    } 

    if ( gg == 0.0 ) /*Unlikely. If gradient is exactly zero then we are already done. */
    { FREEALL 
      return true; 
    } 

    gam = dgg / gg; 

    for ( j=1; j<=n; j++ ) 
    { g[j] = -xi[j]; 
      xi[j] = h[j] = g[j] + gam * h[j]; 
    }
  } 
  FREEALL
  return false;
}

void linmin(  SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float p[], float xi[], int n, float *fret, 
              float (*func)( SceneData &, const FeaturePair &, int [], int, float [] ) )

/*Given an n-dimensional point p[1..n] and an n-dimensional direction xi[1..n], moves and resets p to where the function func(p) takes 
on a minimum along the direction xi from p, and replaces xi by the actual vector displacement that p was moved. Also returns as fret
 the value of func at the returned location p. This is actually all accomplished by calling the routines mnbrak and brent.*/
  
{ float brent( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float ax, float bx, float cx, 
               float (*f)( SceneData &, const FeaturePair &, int [], int, float ), float tol, float *xmin ); 

  float f1dim( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float x ); 
  void mnbrak( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float *ax, float *bx, float *cx, float *fa, float *fb, float *fc, 
               float (*func)( SceneData &, const FeaturePair &, int [], int, float ) ); 
  
  int j; 
  float xx, xmin, fx, fb, fa, bx, ax; 
  ncom = n; 
  pcom = vector( 1, n ); 
  xicom = vector( 1, n ); 
  nrfunc = func; 
  
  for ( j=1; j<=n; j++ ) 
  { pcom[j] = p[j]; 
    xicom[j] = xi[j]; 
  } 
  
  ax = 0.0; 
  xx=10.0; 
  mnbrak( sceneData, pair, imageIDs, imageNumber, &ax, &xx, &bx, &fa, &fx, &fb, f1dim ); 
  *fret = brent( sceneData, pair, imageIDs, imageNumber, ax, xx, bx, f1dim, TOL, &xmin ); 
  
  for ( j=1; j<=n; j++ ) 
  { xi[j] *= xmin; 
    p[j] += xi[j]; 
  } 
  
  free_vector( xicom, 1, n ); 
  free_vector( pcom, 1, n ); 
} 


float f1dim( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float x )
{ int j; 
  float f, *xt; 
  xt = vector( 1, ncom ); 
  
  for ( j=1; j<=ncom; j++ ) 
    xt[j] = pcom[j] + x * xicom[j] ; 
    
  f = (*nrfunc)( sceneData, pair, imageIDs, imageNumber, xt); 
  free_vector( xt, 1, ncom ); 
  return f; 
}



void mnbrak( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float *ax, float *bx, float *cx, float *fa, float *fb, float *fc, 
             float (*func)( SceneData &, const FeaturePair &, int [], int, float ) ) 

/*Given a function func, and given distinct initial points ax and bx, this routine searches in the downhill direction (defined by the
 function as evaluated at the initial points) and returns new points ax, bx, cx that bracket a minimum of the function. Also returned
 are the function values at the three points, fa, fb, and fc.*/
 
{ float ulim, u, r, q, fu, dum; 
  *fa = (*func)( sceneData, pair, imageIDs, imageNumber, *ax); 
  *fb = (*func)( sceneData, pair, imageIDs, imageNumber, *bx);
   
  if ( *fb > *fa ) 
  { SHFT( dum, *ax, *bx, dum ) 
    SHFT( dum, *fb, *fa, dum ) 
  } 
    
  *cx = (*bx) + GOLD * ( *bx - *ax ); 
  *fc = (*func)( sceneData, pair, imageIDs, imageNumber, *cx ); 
  
  while (*fb > *fc) 
  { r=(*bx-*ax)*(*fb-*fc); 
    q=(*bx-*cx)*(*fb-*fa); 
    u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/(2.0*TSIGN(FMAX(fabs(q-r),TINY),q-r)); 
    ulim=(*bx)+GLIMIT*(*cx-*bx); 
    
    if ((*bx-u)*(u-*cx) > 0.0) 
    { fu=(*func)(sceneData, pair, imageIDs, imageNumber, u); 
      
      if (fu < *fc) 
      { *ax=(*bx); 
        *bx=u; 
        *fa=(*fb); 
        *fb=fu; 
        return; 
      } 
      
      else if (fu > *fb) 
      { *cx=u; 
        *fc=fu; 
        return; 
      }
       
      u=(*cx)+GOLD*(*cx-*bx); 
      fu=(*func)(sceneData, pair, imageIDs, imageNumber, u); 
    } 
    
    else if ((*cx-u)*(u-ulim) > 0.0) 
    { fu=(*func)(sceneData, pair, imageIDs, imageNumber, u); 
      
      if (fu < *fc) 
      { SHFT(*bx,*cx,u,*cx+GOLD*(*cx-*bx)) 
        SHFT(*fb,*fc,fu,(*func)(sceneData, pair, imageIDs, imageNumber, u)) 
      } 
    } 
    
    else if ((u-ulim)*(ulim-*cx) >= 0.0) 
    { u=ulim; 
      fu=(*func)(sceneData, pair, imageIDs, imageNumber, u); 
    } 
    
    else 
    { u=(*cx)+GOLD*(*cx-*bx); 
      fu=(*func)(sceneData, pair, imageIDs, imageNumber, u); 
    } 
    
    SHFT(*ax,*bx,*cx,u) 
    SHFT(*fa,*fb,*fc,fu) 
  } 
}


  
float brent( SceneData & sceneData, const FeaturePair & pair, int imageIDs[], int imageNumber, float ax, float bx, float cx, 
             float (*f)( SceneData &, const FeaturePair &,int [], int, float ), float tol, float *xmin )

/*Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is between ax and cx, and f(bx) is less than 
both f(ax) and f(cx)), this routine isolates the minimum to a fractional precision of about tol using Brent s method. The abscissa of 
the minimum is returned as xmin, and the minimum function value is returned as brent, the returned function value.*/

{ int iter; 
  float a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm; 
  float e=0.0; 
  a=(ax < cx ? ax : cx); 
  b=(ax > cx ? ax : cx); 
  x=w=v=bx; 
  fw=fv=fx=(*f)(sceneData, pair, imageIDs, imageNumber, x); 
  
  for (iter=1;iter<=ITMAX;iter++) 
  { xm=0.5*(a+b); 
    tol2=2.0*(tol1=tol*fabs(x)+ZEPS); 
    
    if (fabs(x-xm) <= (tol2-0.5*(b-a))) 
    { *xmin=x; 
      return fx; 
    } 
    
    if (fabs(e) > tol1) 
    { r=(x-w)*(fx-fv); 
      q=(x-v)*(fx-fw); 
      p=(x-v)*q-(x-w)*r; 
      q=2.0*(q-r);
       
      if (q > 0.0) 
        p = -p; 
        
      q=fabs(q);
      etemp=e; 
      e=d; 
      
      if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x)) 
        d=CGOLD*(e=(x >= xm ? a-x : b-x));
      
      else 
      { d=p/q; 
        u=x+d; 
        
        if (u-a < tol2 || b-u < tol2) 
          d=TSIGN(tol1,xm-x); 
      } 
    } 
    
    else 
    { d=CGOLD*(e=(x >= xm ? a-x : b-x)); 
    } 
    
    u=(fabs(d) >= tol1 ? x+d : x+TSIGN(tol1,d)); 
    fu=(*f)(sceneData, pair, imageIDs, imageNumber, u); 
    
    if (fu <= fx) 
    { if (u >= x) a=x; 
      else b=x; 
      SHFT(v,w,x,u) 
      SHFT(fv,fw,fx,fu) 
    } 
    
    else 
    { if (u < x) a=u; 
      else b=u; 
      
      if (fu <= fw || w == x) 
      { v=w; 
        w=u; 
        fv=fw; 
        fw=fu; 
      } 
      
      else if (fu <= fv || v == x || v == w) 
      { v=u; 
        fv=fu; 
      } 
    } 
  } 
  *xmin=x; 
  return fx; 
}
