#include "Disable4786.hpp"
#include "IntegerSet.hpp"
#include <utility>
#include <numeric>
#include <iostream>

using namespace std;

namespace {

  template<typename T> inline T min( T a, T b )
  {
    return a < b ? a : b;
  }

  template<typename T> inline T max( T a, T b )
  {
    return a > b ? a : b;
  }

}


bool IntegerSet::CompareIntegerRange::operator() ( 
  const IntegerRange & integerRange0, const IntegerRange & integerRange1 ) const
{
  return integerRange0.GetUpperBound() < integerRange1.GetLowerBound();
}

IntegerSet::IntegerSet()
{}

IntegerSet::IntegerSet( const IntegerSet & integerSet ):
  ranges_( integerSet.ranges_ )
{}

IntegerSet::IntegerSet( const IntegerRange & integerRange )
{
  if( !integerRange.IsEmpty() )
    ranges_.insert( integerRange );
}

const IntegerSet & IntegerSet::operator=( const IntegerSet & integerSet )
{
  ranges_ = integerSet.ranges_;
  return *this;
}

const IntegerSet & IntegerSet::operator+=( const IntegerSet & integerSet )
{
  typedef IntegerRangeSet::iterator Iter;
  typedef IntegerRangeSet::const_iterator ConstIter;

  ConstIter newRange = integerSet.ranges_.begin();
  ConstIter lastNewRange = integerSet.ranges_.end();
  for( ; newRange != lastNewRange; ++newRange ) {
    //
    // Insert the current range.
    //
    pair<Iter, Iter> equalRange( ranges_.equal_range( *newRange ) );
    if( equalRange.first != equalRange.second ) {
      Iter lastEqual = equalRange.second;
      --lastEqual;
      int lowerBound = min( newRange->GetLowerBound(), 
				                equalRange.first->GetLowerBound() );
      int upperBound = max( newRange->GetUpperBound(), 
				                lastEqual->GetUpperBound() );
      ranges_.erase( equalRange.first, equalRange.second );
      ranges_.insert( IntegerRange( lowerBound, upperBound ) );
    }
    else {
      ranges_.insert( *newRange );
    }
  }

  return *this;
}

const IntegerSet & IntegerSet::operator-=( const IntegerSet & integerSet )
{
  typedef IntegerRangeSet::iterator Iter;
  typedef IntegerRangeSet::const_iterator ConstIter;

  ConstIter newRange = integerSet.ranges_.begin();
  ConstIter lastNewRange = integerSet.ranges_.end();
  for( ; newRange != lastNewRange; ++newRange ) {
    //
    // Erase the current range.
    //
    pair<Iter, Iter> equalRange( ranges_.equal_range( *newRange ) );
    if( equalRange.first != equalRange.second ) {
      Iter lastEqual = equalRange.second;
      --lastEqual;
      int lowerBound0 = equalRange.first->GetLowerBound();
      int upperBound0 = newRange->GetLowerBound();
      int lowerBound1 = newRange->GetUpperBound();
      int upperBound1 = lastEqual->GetUpperBound();
      ranges_.erase( equalRange.first, equalRange.second );
      if( lowerBound0 < upperBound0 )
        ranges_.insert( IntegerRange( lowerBound0, upperBound0 ) );
      if( lowerBound1 < upperBound1 )
        ranges_.insert( IntegerRange( lowerBound1, upperBound1 ) );
    }
  }

  return *this;
}

const IntegerSet & IntegerSet::operator&=( const IntegerSet & integerSet )
{
  typedef IntegerRangeSet::iterator Iter;
  typedef IntegerRangeSet::const_iterator ConstIter;

  IntegerRangeSet newRanges;

  ConstIter newRange = integerSet.ranges_.begin();
  ConstIter lastNewRange = integerSet.ranges_.end();
  for( ; newRange != lastNewRange; ++newRange ) {
    pair<Iter, Iter> equalRange( ranges_.equal_range( *newRange ) );
    for( ; equalRange.first != equalRange.second; ++equalRange.first ) {
      IntegerRange intersection( Intersect( *(equalRange.first), *newRange ) );
      if( !intersection.IsEmpty() )
        newRanges.insert( intersection );
    }
  }

  ranges_.swap( newRanges );

  return *this;
}

bool IntegerSet::operator==( const IntegerSet & integerSet ) const
{
  return ranges_.size() == integerSet.ranges_.size() && 
    equal( ranges_.begin(), ranges_.end(), integerSet.ranges_.begin() );
}

bool IntegerSet::operator!=( const IntegerSet & integerSet ) const
{
  return ranges_.size() != integerSet.ranges_.size() ||
    !equal( ranges_.begin(), ranges_.end(), integerSet.ranges_.begin() );
}

int IntegerSet::GetSize() const
{
  typedef IntegerRangeSet::const_iterator ConstIter;

  ConstIter range, lastRange;
  range = ranges_.begin();
  lastRange = ranges_.end();
  int size = 0;
  for( ; range != lastRange; ++range )
    size += range->GetSize();

  return size;
}

const IntegerSet operator+( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 )
{
  IntegerSet result( integerSet0 );
  return result += integerSet1;
}

const IntegerSet operator-( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 )
{
  IntegerSet result( integerSet0 );
  return result -= integerSet1;
}

const IntegerSet operator&( const IntegerSet & integerSet0, 
		const IntegerSet & integerSet1 )
{
  IntegerSet result( integerSet0 );
  return result &= integerSet1;
}

ostream & operator<<( ostream & os, const IntegerSet & integerSet )
{
  os << "{ ";
  copy( integerSet.ranges_.begin(), integerSet.ranges_.end(), ostream_iterator<IntegerRange>( os, " " ) );
  os << "}";
  return os;
}
