#include "NonanchorReconstructor.hpp"
#include <iostream>
#include "nrutil.h"

using std::cout;
using std::endl;

void svbksb( float **u, float w[], float **v, int m, int n, float b[], float x[] );
void svdcmp( float **a, int m, int n, float w[], float **v );


/*****************************************************************************
*****************************************************************************/
NonanchorReconstructor::NonanchorReconstructor( SceneData & sceneData ): 
  sceneData_( sceneData )
{}


/*****************************************************************************
*****************************************************************************/
NonanchorReconstructor::NonanchorReconstructor( 
		const NonanchorReconstructor & nonanchorReconstructor ): 
      sceneData_( nonanchorReconstructor.sceneData_ )
{}


/*****************************************************************************
*****************************************************************************/
void NonanchorReconstructor::LinearEquationsSolver( float **A, float *x, 
		float *B, int m, int n, float & conditionNumber )
{ float **u = matrix(1, m, 1, n);
  float **v = matrix(1, n, 1, n);
  float *w = vector(1, n);
  float *b = vector(1, m);
  int i,j;
  float wmax, wmin;

  for (i = 1; i <= n; ++i) 
    for (j = 1; j <= m; ++j) 
      u[j][i] = A[j][i];

  for (j = 1; j <= m; ++j) 
    b[j] = B[j];

  for (i = 1; i <= n; ++i) 
    x[i] = 0.0;

  // do a singular value decomposition
  svdcmp(u, m, n, w, v);

  wmax=0.0; 

  for ( j=1; j<=n; j++ ) 
    if ( w[j] > wmax ) 
      wmax=w[j]; 

  conditionNumber = 0.0;

  for ( j=1; j<=n; j++ )
    conditionNumber += wmax/w[j];

  wmin=wmax*1.0e-6; 

  for( j=1; j<=n; j++ ) 
    if ( w[j] < wmin ) 
      w[j]=0.0;  

  // do a backsubstitution
  svbksb(u, w, v, m, n, b, x);

  free_matrix( u, 1, m, 1, n );
  free_matrix( v, 1, n, 1, n );
  free_vector( w, 1, n );
  free_vector( b, 1, m );
}


/*****************************************************************************
*****************************************************************************/
void NonanchorReconstructor::ReconstructNonanchorFeature( 
		const Feature & feature, float & projError, int nonanchorImageNumber )
{ 
  VM3Vector *C, *Q[2], *M[3], *coord[3];
  VM3Vector averageC( 0.0, 0.0, 0.0 ), reconLocation;
  int i, j, Inum, *image, lowerBound;
  float *gamma[2], *k[3], conditionNumber, viewpointVariance;
  float meanViewpointDistance, scale=0;

  Inum = feature.GetRange().GetSize();
  
  if ( Inum < nonanchorImageNumber )
    return;

  //cout << "Reconstructing Nonanchor: " << feature.GetID() << endl;
  
  lowerBound = feature.GetRange().GetLowerBound();  
  std::vector<int> validImage( Inum );

  for ( i=0, j=0; i<Inum; i++ )
  { if( ( sceneData_.GetViewpointValidOrientationList()[lowerBound+i] ) && 
		  ( sceneData_.GetAnchorList()[lowerBound+i] ) )
	{ validImage[j] = lowerBound + i;
	  j++;
	}
  }

  validImage.resize(j);

  if ( j < nonanchorImageNumber )
    return;
  
  Inum = j / nonanchorImageNumber;

  image = (int *) malloc( nonanchorImageNumber * sizeof( int ) );

  for ( i=0; i<nonanchorImageNumber; i++ )  
    image[i] = validImage[i * Inum];	  	  
 
  Inum = nonanchorImageNumber;

  C = (VM3Vector *) malloc( nonanchorImageNumber * sizeof( VM3Vector ) );

  for ( i=0; i<2; i++ )
  { Q[i] = (VM3Vector *) malloc( nonanchorImageNumber * sizeof( VM3Vector ) );
    gamma[i] = (float *) malloc( nonanchorImageNumber * sizeof( float ) );
  }

  for ( i=0; i<3; i++ )
  { M[i] = (VM3Vector *) malloc( nonanchorImageNumber * sizeof( VM3Vector ) );
    coord[i] = (VM3Vector *) malloc( nonanchorImageNumber * sizeof( VM3Vector ) );
    k[i] = (float *) malloc( nonanchorImageNumber * sizeof( float ) );
  }

  float **A = matrix(1, 3*nonanchorImageNumber, 1, 3+nonanchorImageNumber);
  float *x = vector(1, 3+nonanchorImageNumber);
  float *B = vector(1, 3*nonanchorImageNumber); 

  for ( i=0; i<Inum; i++ ) {
    C[i] = sceneData_.GetViewpointLocationList()[image[i]];
  }

  for ( i=0; i<Inum; i++ )
    averageC += C[i];

  averageC = averageC / Inum;
  viewpointVariance = 0;

  for ( i=0; i<Inum; i++ )
    viewpointVariance += ( C[i] - averageC ).lengthSquared();

  viewpointVariance = viewpointVariance / ( Inum - 1 );
  
  for ( i=0; i<Inum; i++ )
  { Q[0][i]=sceneData_.GetReconObjectList()[sceneData_.GetAnchorList()[image[i]]->GetFirstFeature().GetID()].GetReconLocation();
    Q[1][i]=sceneData_.GetReconObjectList()[sceneData_.GetAnchorList()[image[i]]->GetSecondFeature().GetID()].GetReconLocation();
  }

  for ( i=0; i<Inum; i++ )
  { scale += ( Q[0][i] - C[i] ).length();
    scale += ( Q[1][i] - C[i] ).length();
  }

  scale /= float( 2 * Inum );

  for ( i=0; i<Inum; i++ )
  { C[i] = C[i] / scale;
    Q[0][i] = Q[0][i] / scale;
	Q[1][i] = Q[1][i] / scale;
  }

  for ( i=0; i<Inum; i++ )
  { M[0][i]=Q[0][i]-C[i];
    M[1][i]=(Q[0][i]-C[i]).dot(Q[0][i]-C[i])*(Q[1][i]-C[i])-
		 (Q[0][i]-C[i]).dot(Q[1][i]-C[i])*(Q[0][i]-C[i]);
    M[2][i]=(Q[0][i]-C[i]).cross(Q[1][i]-C[i]);
  }    

  for ( i=0; i<Inum; i++ )
  { coord[0][i].set( sceneData_.GetAnchorList()[image[i]]->GetFirstFeature().GetLocation(image[i]).x_ * sceneData_.GetCameraParameters().realLength,
                     sceneData_.GetAnchorList()[image[i]]->GetFirstFeature().GetLocation(image[i]).y_ * sceneData_.GetCameraParameters().realWidth,
                     -sceneData_.GetCameraParameters().focus );
    coord[1][i].set( sceneData_.GetAnchorList()[image[i]]->GetSecondFeature().GetLocation(image[i]).x_ * sceneData_.GetCameraParameters().realLength,
                     sceneData_.GetAnchorList()[image[i]]->GetSecondFeature().GetLocation(image[i]).y_ * sceneData_.GetCameraParameters().realWidth,
                     -sceneData_.GetCameraParameters().focus );
    coord[2][i].set( feature.GetLocation(image[i]).x_ * sceneData_.GetCameraParameters().realLength,
                     feature.GetLocation(image[i]).y_ * sceneData_.GetCameraParameters().realWidth,
                     -sceneData_.GetCameraParameters().focus );
  }

#if 0
  for( i=0; i<Inum; ++i ) {
    cout << " Image " << image[i] << ":" << endl
      << "    CoP: (" << C[i].x() << ", " << C[i].y() << ", " << C[i].z() << ")" << endl
      << "    A1:  (" << Q[0][i].x() << ", " << Q[0][i].y() << ", " << Q[0][i].z() << ")" << endl
      << "    A2:  (" << Q[1][i].x() << ", " << Q[1][i].y() << ", " << Q[1][i].z() << ")" << endl
      << "    PA1: (" << coord[0][i].x() << ", " << coord[0][i].y() << ", " << coord[0][i].z() << ")" << endl
      << "    PA2: (" << coord[1][i].x() << ", " << coord[1][i].y() << ", " << coord[1][i].z() << ")" << endl
      << "    PN:  (" << coord[2][i].x() << ", " << coord[2][i].y() << ", " << coord[2][i].z() << ")" << endl
      ;
  }
#endif

  for ( i=0; i<2; i++ )
  {  for ( j=0; j<Inum; j++ )
     { gamma[i][j]=
		  sqrt((Q[i][j]-C[j]).dot(Q[i][j]-C[j])/coord[i][j].dot(coord[i][j]));
     }
  }

  for ( i=0; i<Inum; i++ )
  { k[0][i]=
	    coord[0][i].dot(coord[2][i]);
    k[1][i]=
		 (coord[0][i].cross(coord[2][i])).dot(coord[0][i].cross(coord[1][i]));
    k[2][i]=
		 coord[2][i].dot(coord[0][i].cross(coord[1][i]));
  }

  for ( i=1; i<=3*Inum; i++ )
    for ( j=1; j<=3+Inum; j++ )
      A[i][j]=0;

  for ( i=0; i<Inum; i++ )
  { for ( j=0; j<3; j++ )
    { A[1+j+3*i][1]=M[j][i].x();
      A[1+j+3*i][2]=M[j][i].y();
      A[1+j+3*i][3]=M[j][i].z();
    }
    A[1+3*i][4+i]=-gamma[0][i]*k[0][i];
    A[2+3*i][4+i]=-gamma[0][i]*gamma[0][i]*gamma[1][i]*k[1][i];
    A[3+3*i][4+i]=-gamma[0][i]*gamma[1][i]*k[2][i];
  }

  for ( i=0; i<Inum; i++ )
    for ( j=0; j<3; j++ )
      B[1+j+3*i]=C[i].dot(M[j][i]);

  LinearEquationsSolver( A, x, B, 3*Inum, 3+Inum, conditionNumber );

  reconLocation.set( x[1], x[2], x[3] );
  reconLocation = scale * reconLocation;
  //cout << " Nonanchor " << feature.GetID() << ": " << x[1] << " " << x[2] << " " << x[3] << endl;

  meanViewpointDistance = 0;

  for ( i=0; i<Inum; i++ )
    meanViewpointDistance += ( reconLocation - C[i] ).length();

  meanViewpointDistance /= Inum;

  ReconObject & reconObject = sceneData_.GetReconObjectList()[feature.GetID()];
  reconObject.SetNonanchorReconstruction( reconLocation, conditionNumber, 
		  viewpointVariance, meanViewpointDistance );

  //Projection Error knowing the orientation
  /*VM3Vector coord3;
  projError = 0;

  for ( i=0; i<validImage.size(); i++ )
  { VM4Matrix invertOrientation( sceneData_.GetViewpointOrientationList()[validImage[i]] );

    invertOrientation.invertAnglePreserving();
    coord3 = invertOrientation * ( reconLocation - sceneData_.GetViewpointLocationList()[validImage[i]] );
    
	float z = coord3.z();
    coord3 = coord3 / z * ( -sceneData_.GetCameraParameters().focus );
    projError += SQRT( ( coord3.x() / sceneData_.GetCameraParameters().realLength - feature.GetLocation(validImage[i]).x_ ) * 
                       ( coord3.x() / sceneData_.GetCameraParameters().realLength - feature.GetLocation(validImage[i]).x_ ) +
                       ( coord3.y() / sceneData_.GetCameraParameters().realWidth - feature.GetLocation(validImage[i]).y_ ) * 
                       ( coord3.y() / sceneData_.GetCameraParameters().realWidth - feature.GetLocation(validImage[i]).y_ ) );
  }

  projError /= validImage.size();*/

  /*projError = 0;
  VM3Vector coord2[3];

  float realLength = sceneData_.GetCameraParameters().realLength;
  float realWidth = sceneData_.GetCameraParameters().realWidth;
  float focus = -sceneData_.GetCameraParameters().focus;

  for ( i=0; i<validImage.size(); i++ )
  { Feature firstFeature = sceneData_.GetAnchorList()[validImage[i]]->GetFirstFeature();
    Feature secondFeature = sceneData_.GetAnchorList()[validImage[i]]->GetSecondFeature();
	
	coord2[0].set( firstFeature.GetLocation(validImage[i]).x_ * realLength,
                   firstFeature.GetLocation(validImage[i]).y_ * realWidth,
                   -focus );
    coord2[1].set( secondFeature.GetLocation(validImage[i]).x_ * realLength,
                   secondFeature.GetLocation(validImage[i]).y_ * realWidth,
                   -focus );
    coord2[2].set( feature.GetLocation(validImage[i]).x_ * realLength,
                   feature.GetLocation(validImage[i]).y_ * realWidth,
                   -focus );

	float C_a1 = coord2[0].length();
	float C_a2 = coord2[1].length();
	float C_q = coord2[2].length();
	float a1_q = ( coord2[0] - coord2[2] ).length();
	float a2_q = ( coord2[1] - coord2[2] ).length();

	float cos1 = coord2[0].dot( coord2[0] - coord2[2] )/( C_a1 * a1_q );
	float sin1 = sqrt( 1 - cos1 * cos1 );

	VM3Vector C_A1 = sceneData_.GetReconObjectList()[firstFeature.GetID()].GetReconLocation() - sceneData_.GetViewpointLocationList()[validImage[i]];
	VM3Vector C_A2 = sceneData_.GetReconObjectList()[secondFeature.GetID()].GetReconLocation() - sceneData_.GetViewpointLocationList()[validImage[i]];
	VM3Vector C_Q = sceneData_.GetReconObjectList()[feature.GetID()].GetReconLocation() - sceneData_.GetViewpointLocationList()[validImage[i]];
	  
	float cos2 = C_A1.dot( C_Q )/( C_A1.length() * C_Q.length() );
	float sin2 = sqrt( 1 - cos2 * cos2 );
	  
	float sin3 = sin1 * cos2 + cos1 * sin2;
	float a1_qt = C_a1 * sin2 / sin3;

	float tempError = ( ABS( a1_qt - a1_q ) + estimatedTrackingError * realLength ) * C_Q.length() / C_q * sin3;
      
	//if ( projError < tempError )
	projError += tempError;

	cos1 = coord2[1].dot( coord2[1] - coord2[2] )/( C_a2 * a2_q );
	sin1 = sqrt( 1 - cos1 * cos1 );

	cos2 = C_A2.dot( C_Q )/( C_A2.length() * C_Q.length() );
    sin2 = sqrt( 1 - cos2 * cos2 );
 
	sin3 = sin1 * cos2 + cos1 * sin2;
	float a2_qt = C_a2 * sin2 / sin3;

	tempError = ( ABS( a2_qt - a2_q ) + estimatedTrackingError * realLength ) * C_Q.length() / C_q * sin3;

  	//if ( projError < tempError )
	projError += tempError;
  }

  float cosAngleTemp, cosAngle = 1;

  for ( i=0; i<Inum; i++ )
  { for ( j=i+1; j<Inum; j++ )
    { cosAngleTemp = coord[2][i].dot( coord[2][j] ) / ( coord[2][i].length() * coord[2][j].length() );
      if ( ABS( cosAngle ) > ABS( cosAngleTemp ) )
      {  cosAngle = cosAngleTemp;
	  }
    }
  }

  float sinAngle = sqrt( ( 1 - cosAngle ) / 2 );

  projError /= ( 2 * validImage.size() * sinAngle );*/

  free( C );
  free( image );

  for ( i=0; i<2; i++ )
  { free( Q[i] );
    free( gamma[i] );
  }

  for ( i=0; i<3; i++ )
  { free( M[i] );
    free( coord[i] );
    free( k[i] );
  }

  free_matrix( A, 1, 3*nonanchorImageNumber, 1, 3+nonanchorImageNumber );
  free_vector( x, 1, 3+nonanchorImageNumber );
  free_vector( B, 1, 3*nonanchorImageNumber ); 
}
