#ifndef __Vector3_hpp__
#define __Vector3_hpp__

#include <cstdlib>

///
/// @brief A 3-dimensional vector.
///
class Vector3 {

  float data_[4];                     ///< The vector data.

public:

  ///
  /// @brief The zero vector.
  ///
  static const Vector3 ZERO;

  ///
  /// @brief The three unit vectors.
  ///
  static const Vector3 UNIT[3];

  ///
  /// @brief Construct an uninitialized vector.
  ///
  Vector3();

  ///
  /// @brief Construct a vector from its components.
  ///
  Vector3( float x, float y, float z );

  ///
  /// @brief Component access.
  ///
  float & operator[]( size_t i );

  ///
  /// @brief Component access.
  ///
  float operator[]( size_t i ) const;

  ///
  /// @brief Negation operator.
  ///
  const Vector3 operator-() const;

  ///
  /// @name Arithmetic operators.
  ///
  /// @{
  const Vector3 & operator*=( float s );
  const Vector3 & operator/=( float s );
  const Vector3 & operator+=( const Vector3 & v );
  const Vector3 & operator-=( const Vector3 & v );
  /// @}
  ///

  ///
  /// @brief Compute the squared length of the vector.
  ///
  float GetSquaredLength() const;

  ///
  /// @brief Compute the length of the vector.
  ///
  float GetLength() const;

};

///
/// @name Vector arithmetic operators.
///
/// @{
const Vector3 operator*( const Vector3 & v, float s );
const Vector3 operator*( float s, const Vector3 & v );
const Vector3 operator/( const Vector3 & v, float s );
const Vector3 operator+( const Vector3 & v1, const Vector3 & v2 );
const Vector3 operator-( const Vector3 & v1, const Vector3 & v2 );
/// @}
///

///
/// @brief Normalize a Vector3.
///
void Normalize( Vector3 & v );

///
/// @brief Compute the dot product of two Vector3's.
///
float Dot( const Vector3 & v1, const Vector3 & v2 );

///
/// @brief Compute the cross product of two Vector3's.
///
const Vector3 Cross( const Vector3 & v1, const Vector3 & v2 );

#include "Vector3.inl"

#endif // __Vector3_hpp__
