#include <common.h>

#define NPARMS 800


extern "C" int lmdif_(int (*)(int *, int *, double *, double *), 
                      int *m, int *n, double *x, double *fvec, 
                      double *ftol, double *xtol, double *gtol, 
                      int *maxfev, double *epsfcn, double *diag, 
                      int *mode, double *factor, int *nprint, 
                      int *info, int *nfev, double *fjac, 
                      int *ldfjac, int *ipvt, double *qtf, 
                      double *wa1, double *wa2, double *wa3, 
                      double *wa4);


/*****************************************************************************
*****************************************************************************/
static int
lmdifError (int *m_ptr, int *n_ptr, double *parms, double *error)
{
   for (int n=0; n<NPARMS; n++) {
      error[n] = SQR(parms[n]-n);
      //printf("%lf ", parms[n]);
   }
   //printf("\n");

   return (TRUE);
}


/*****************************************************************************
*****************************************************************************/
/* Parameters controlling MINPACK's lmdif() optimization routine. */
/* See the file lmdif.f for definitions of each parameter.        */
#define REL_SENSOR_TOLERANCE_ftol    1.0E-5      /* [pix] */
#define REL_PARAM_TOLERANCE_xtol     1.0E-7
#define ORTHO_TOLERANCE_gtol         0.0
#define MAXFEV                       (1000*n)
#define EPSFCN                       1.0E-10 /* was E-16 Do not set to 0! */
//#define EPSFCN                       1.0E-4 /* was E-16 Do not set to 0! */
#define MODE                         1       /* variables scaled internally */
#define FACTOR                       100.0 
static void
lmdifSolve (double *parms, int nparms, int nerrors)
{
    /* Parameters needed by MINPACK's lmdif() */
    int     n = nparms;
    int     m = nerrors;
    double *x;
    double *fvec;
    double  ftol = REL_SENSOR_TOLERANCE_ftol;
    double  xtol = REL_PARAM_TOLERANCE_xtol;
    double  gtol = ORTHO_TOLERANCE_gtol;
    int     maxfev = MAXFEV;
    double  epsfcn = EPSFCN;
    double *diag;
    int     mode = MODE;
    double  factor = FACTOR;
    int     nprint = 0;
    int     info;
    int     nfev;
    double *fjac;
    int     ldfjac = m;
    int    *ipvt;
    double *qtf;
    double *wa1;
    double *wa2;
    double *wa3;
    double *wa4;


    /* allocate stuff dependent on n */
    x    = (double *)calloc(n, sizeof(double));
    diag = (double *)calloc(n, sizeof(double));
    qtf  = (double *)calloc(n, sizeof(double));
    wa1  = (double *)calloc(n, sizeof(double));
    wa2  = (double *)calloc(n, sizeof(double));
    wa3  = (double *)calloc(n, sizeof(double));
    ipvt = (int    *)calloc(n, sizeof(int));

    /* allocate some workspace */
    if (( fvec = (double *) calloc ((unsigned int) m, 
                                    (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fvec\n");
       exit(-1);
    }

    if (( fjac = (double *) calloc ((unsigned int) m*n,
                                    (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace fjac\n");
       exit(-1);
    }

    if (( wa4 = (double *) calloc ((unsigned int) m, 
                                   (unsigned int) sizeof(double))) == NULL ) {
       fprintf(stderr,"calloc: Cannot allocate workspace wa4\n");
       exit(-1);
    }


    /* copy parameters in as initial values */
    for (int i=0; i<n; i++)
       x[i] = parms[i];

    /* define optional scale factors for the parameters */
    if ( mode == 2 ) {
        for (i = 0; i < n; i++)
           diag[i] = 1.0;             /* some user-defined values */
    }

    /* perform the optimization */ 
    printf("Starting optimization step...\n");
    lmdif_ (lmdifError,
            &m, &n, x, fvec, &ftol, &xtol, &gtol, &maxfev, &epsfcn,
            diag, &mode, &factor, &nprint, &info, &nfev, fjac, &ldfjac,
            ipvt, qtf, wa1, wa2, wa3, wa4);
    double totalerror = 0;
    for (i=0; i<m; i++)
       totalerror += fvec[i];
    printf("\tnum function calls = %i\n", nfev);
    printf("\tremaining total error value = %f\n", totalerror);
    printf("...ended optimization step.\n");

    /* copy result back to parameters array */
    for (i=0; i<n; i++)
       parms[i] = x[i];

    /* release allocated workspace */
    free (fvec);
    free (fjac);
    free (wa4);
    free (ipvt);
    free (wa1);
    free (wa2);
    free (wa3);
    free (qtf);
    free (diag);
    free (x);
}


/*****************************************************************************
*****************************************************************************/
int
main()
{
   int nerrors = NPARMS;
   int nparms = NPARMS;
   double parms[NPARMS];

   for (int n=0; n<NPARMS; n++) 
      parms[n] = n * 0.1;

   lmdifSolve(parms, nparms, nerrors);

   printf("result:\n");
   for (n=0; n<NPARMS; n++) 
      printf("%lf ", parms[n]);
   printf("\n");


   return (TRUE);
}

