#include "cmatrix.h"


#define SMALL 1e-8              /* Small enough to be considered zero */

/*****************************************************************************

Shuffles rows in inverse of 3x3.  See comment in CMatrix_inv3_second_col().

*****************************************************************************/
static void
matrix_inv3_swap(float inv[3][3], int row0, int row1, int row2)
{
   register int i, tempi;
   float       temp;

#define SWAP_ROWS(a, b) \
   for (i = 0; i < 3; i++) SWAP(inv[a][i], inv[b][i], temp); \
   SWAP(a, b, tempi)

   if (row0 != 0){
      if (row1 == 0) {
         SWAP_ROWS(row0, row1);
      }
      else {
         SWAP_ROWS(row0, row2);
      }
   }

   if (row1 != 1) {
      SWAP_ROWS(row1, row2);
   }
}


/*****************************************************************************

Does Gaussian elimination on second column.

*****************************************************************************/
static int
matrix_inv3_second_col(float source[3][3], float inv[3][3], int row0)
{
   register int row1, row2, i1, i2, i;
   float       temp;
   float       a, b;

   /* Find which row to use */
   if      (row0 == 0)  i1 = 1, i2 = 2;
   else if (row0 == 1)  i1 = 0, i2 = 2;
   else                 i1 = 0, i2 = 1;

   /* Find which is larger in abs. val.:the entry in [i1][1] or [i2][1] */
   /* and use that value for pivoting.                                  */

   a = source[i1][1]; if (a < 0) a = -a;
   b = source[i2][1]; if (b < 0) b = -b;
   if (a > b)   row1 = i1;
   else         row1 = i2;
   row2 = (row1 == i1 ? i2 : i1);

   /* Scale row1 in source */
   if ((source[row1][1] < SMALL) && (source[row1][1] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row1][1];
   source[row1][1]  = 1.0;
   source[row1][2] *= temp;     /* source[row1][0] is zero already */

   /* Scale row1 in inv */
   inv[row1][row1]  = temp;     /* it used to be a 1.0 */
   inv[row1][row0] *= temp;

   /* Clear column one, source, and make corresponding changes in inv */

   for (i = 0; i < 3; i++) if (i != row1) {     /* for i = all rows but row1 */
   temp = -source[i][1];
      source[i][1]  = 0.0;
      source[i][2] += temp * source[row1][2];

      inv[i][row1]  = temp * inv[row1][row1];
      inv[i][row0] += temp * inv[row1][row0];
   }

   /* Scale row2 in source */
   if ((source[row2][2] < SMALL) && (source[row2][2] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row2][2];
   source[row2][2] = 1.0;       /* source[row2][*] is zero already */

   /* Scale row2 in inv */
   inv[row2][row2]  = temp;     /* it used to be a 1.0 */
   inv[row2][row0] *= temp;
   inv[row2][row1] *= temp;

   /* Clear column one, source, and make corresponding changes in inv */
   for (i = 0; i < 3; i++) if (i != row2) {     /* for i = all rows but row2 */
   temp = -source[i][2];
      source[i][2] = 0.0;
      inv[i][row0] += temp * inv[row2][row0];
      inv[i][row1] += temp * inv[row2][row1];
      inv[i][row2] += temp * inv[row2][row2];
   }

   /*
    * Now all is done except that the inverse needs to have its rows shuffled.
    * row0 needs to be moved to inv[0][*], row1 to inv[1][*], etc.
    *
    * We *didn't* do the swapping before the elimination so that we could more
    * easily keep track of what ops are needed to be done in the inverse.
    */
   matrix_inv3_swap(inv, row0, row1, row2);

   return(TRUE);
}


/*****************************************************************************

Fast inversion routine for 3 x 3 matrices.   - Written by jfh.

This takes 30 multiplies/divides, as opposed to 39 for Cramer's Rule.
The algorithm consists of performing fast gaussian elimination, by never
doing any operations where the result is guaranteed to be zero, or where
one operand is guaranteed to be zero. This is done at the cost of clarity,
alas.

Returns 1 if the inverse was successful, 0 if it failed.

*****************************************************************************/
static int
matrix_invert3(float source[3][3], float inv[3][3])
{
   register int i, row0;
   float       temp;
   float       a, b, c;

   inv[0][0] = inv[1][1] = inv[2][2] = 1.0;
   inv[0][1] = inv[0][2] = inv[1][0] = inv[1][2] = inv[2][0] = inv[2][1] = 0.0;

   /* attempt to find the largest entry in first column to use as pivot */
   a = source[0][0]; if (a < 0) a = -a;
   b = source[1][0]; if (b < 0) b = -b;
   c = source[2][0]; if (c < 0) c = -c;

   if (a > b) {
      if (a > c) row0 = 0;
      else row0 = 2;
   }
   else {
      if (b > c) row0 = 1;
      else row0 = 2;
   }

   /* Scale row0 of source */
   if ((source[row0][0] < SMALL) && (source[row0][0] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row0][0];
   source[row0][0]  = 1.0;
   source[row0][1] *= temp;
   source[row0][2] *= temp;

   /* Scale row0 of inverse */
   inv[row0][row0] = temp;      /* other entries are zero -- no effort  */

   /* Clear column zero of source, and make corresponding changes in inverse */

   for (i = 0; i < 3; i++) if (i != row0) {     /* for i = all rows but row0 */
      temp = -source[i][0];
      source[i][0]  = 0.0;
      source[i][1] += temp * source[row0][1];
      source[i][2] += temp * source[row0][2];
      inv[i][row0]  = temp * inv[row0][row0];
   }


   /*
    * We've now done gaussian elimination so that the source and
    * inverse look like this:
    *
    *   1 * *           * 0 0
    *   0 * *           * 1 0
    *   0 * *           * 0 1
    *
    * We now proceed to do elimination on the second column.
    */
   if (!matrix_inv3_second_col(source, inv, row0)) return(FALSE);

   return(TRUE);
}


/*****************************************************************************

Finds a new pivot for a non-simple 4x4.  See comments in MatInvert().

*****************************************************************************/
static int
matrix_inv4_pivot(float src[4][4], float r[4], float *s, int *swap)
{
   register int i, j;
   float        temp, max;

   *swap = -1;

   if (ABS(src[3][3]) < SMALL) {

      /* Look for a different pivot element: one with largest abs value */
      max = 0.0;

      for (i = 0; i < 4; i++) {
         if      (src[i][3] >  max) max =  src[*swap = i][3];
         else if (src[i][3] < -max) max = -src[*swap = i][3];
      }

      /* No pivot element available ! */
      if (*swap < 0) return(FALSE);

      else for (j = 0; j < 4; j++) SWAP(src[*swap][j], src[3][j], temp);
   }

   r[0] = -src[0][3];
   r[1] = -src[1][3];
   r[2] = -src[2][3];

   *s = 1.0 / src[3][3];

   src[0][3] = src[1][3] = src[2][3] = 0.0;
   src[3][3]                         = 1.0;

   src[3][0] *= (*s);
   src[3][1] *= (*s);
   src[3][2] *= (*s);

   for (i = 0; i < 3; i++) {
      src[0][i] += r[0] * src[3][i];
      src[1][i] += r[1] * src[3][i];
      src[2][i] += r[2] * src[3][i];
   }

   return(TRUE);
}


/*****************************************************************************
MatInvert

This returns the inverse of the given matrix.  The result matrix
may be the same as the one to invert.

Fast inversion routine for 4 x 4 matrices, written by jfh.

Returns 1 if the inverse was successful, 0 if it failed.

This routine has been specially tweaked to notice the following:
If the matrix has the form
       * * * 0
       * * * 0
       * * * 0
       * * * 1

(as do many matrices in graphics), then we compute the inverse of
the upper left 3x3 matrix and use this to find the general inverse.

  In the event that the right column is not 0-0-0-1, we do gaussian
elimination to make it so, then use the 3x3 inverse, and then do
our gaussian elimination.

*****************************************************************************/
int
MatInvert(CMatrix result, CMatrix tmat)
{
   CMatrix               mat,inv;
   register int         i, j, simple;
   float                m[3][3], inv3[3][3], s, temp;
   float                r[4], t[4];
   int                  swap;

   /* invert code is row major */
   MatTranspose(mat,tmat);

   MatIdentity(inv);

   /* If last column is not (0,0,0,1), use special code */
   simple = (mat[0][3] == 0.0 && mat[1][3] == 0.0 &&
             mat[2][3] == 0.0 && mat[3][3] == 1.0);

   if (! simple && !matrix_inv4_pivot(mat, r, &s, &swap)) {
      fprintf(stderr, "MatInvert() No inverse!\n");
      return(FALSE);
   }

   t[0] = mat[3][0];
   t[1] = mat[3][1];
   t[2] = mat[3][2];

   /* Copy upper-left 3x3 matrix */
   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) m[i][j] = mat[i][j];

   if (! matrix_invert3(m, inv3)) {
      fprintf(stderr, "MatInvert() No inverse!\n");
      return(FALSE);
   }

   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) inv[i][j] = inv3[i][j];

   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
      inv[3][i] -= t[j] * inv3[j][i];

   if (! simple) {

      /* We still have to undo our gaussian elimination from earlier on */
      /* add r0 * first col to last col */
      /* add r1 * 2nd   col to last col */
      /* add r2 * 3rd   col to last col */

      for (i = 0; i < 4; i++) {
         inv[i][3] += r[0] * inv[i][0] +
                      r[1] * inv[i][1] +
                      r[2] * inv[i][2];
         inv[i][3] *= s;
      }

      if (swap >= 0)
         for (i = 0; i < 4; i++) SWAP(inv[i][swap], inv[i][3], temp);
   }

   /* invert code is row major */
   MatTranspose(result,inv);

   return (TRUE);
}


/*
 * float = det2x2( double a, double b, double c, double d )
 * 
 * calculate the determinent of a 2x2 matrix.
 */
#define det2x2( a,b,c,d )	((a)*(d) - (b)*(c))


/*
 * float = det3x3(  a1, a2, a3, b1, b2, b3, c1, c2, c3 )
 * 
 * calculate the determinent of a 3x3 matrix
 * in the form
 *
 *     | a1,  b1,  c1 |
 *     | a2,  b2,  c2 |
 *     | a3,  b3,  c3 |
 */
static float
det3x3( float a1, float a2, float a3, float b1, float b2, float b3, 
        float c1, float c2, float c3 )
{
    float ans;

    ans = a1 * det2x2( b2, b3, c2, c3 )
        - b1 * det2x2( a2, a3, c2, c3 )
        + c1 * det2x2( a2, a3, b2, b3 );
    return ans;
}


/*
 * float = det4x4( matrix )
 * 
 * calculate the determinent of a 4x4 matrix.
 */
static float
det4x4( CMatrix m )
{
    float ans;
    float a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid selecting */
	/*  correct elements */

	a1 = m[0][0]; b1 = m[0][1]; 
	c1 = m[0][2]; d1 = m[0][3];

	a2 = m[1][0]; b2 = m[1][1]; 
	c2 = m[1][2]; d2 = m[1][3];

	a3 = m[2][0]; b3 = m[2][1]; 
	c3 = m[2][2]; d3 = m[2][3];

	a4 = m[3][0]; b4 = m[3][1]; 
	c4 = m[3][2]; d4 = m[3][3];

    ans = a1 * det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4)
        - b1 * det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4)
        + c1 * det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4)
        - d1 * det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
    return ans;
}



/* 
 *   adjoint( original_matrix, inverse_matrix )
 * 
 *     calculate the adjoint of a 4x4 matrix
 *
 *      Let  a   denote the minor determinant of matrix A obtained by
 *           ij
 *
 *      deleting the ith row and jth column from A.
 *
 *                    i+j
 *     Let  b   = (-1)    a
 *          ij            ji
 *
 *    The matrix B = (b  ) is the adjoint of A
 *                     ij
 */
static void
adjoint( CMatrix in, CMatrix out )
{
    float a1, a2, a3, a4, b1, b2, b3, b4;
    float c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid  */
    /* selecting correct values  */

	a1 = in[0][0]; b1 = in[0][1]; 
	c1 = in[0][2]; d1 = in[0][3];

	a2 = in[1][0]; b2 = in[1][1]; 
	c2 = in[1][2]; d2 = in[1][3];

	a3 = in[2][0]; b3 = in[2][1];
	c3 = in[2][2]; d3 = in[2][3];

	a4 = in[3][0]; b4 = in[3][1]; 
	c4 = in[3][2]; d4 = in[3][3];


    /* row column labeling reversed since we transpose rows & columns */

    out[0][0]  =   det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4);
    out[1][0]  = - det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4);
    out[2][0]  =   det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4);
    out[3][0]  = - det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
        
    out[0][1]  = - det3x3( b1, b3, b4, c1, c3, c4, d1, d3, d4);
    out[1][1]  =   det3x3( a1, a3, a4, c1, c3, c4, d1, d3, d4);
    out[2][1]  = - det3x3( a1, a3, a4, b1, b3, b4, d1, d3, d4);
    out[3][1]  =   det3x3( a1, a3, a4, b1, b3, b4, c1, c3, c4);
        
    out[0][2]  =   det3x3( b1, b2, b4, c1, c2, c4, d1, d2, d4);
    out[1][2]  = - det3x3( a1, a2, a4, c1, c2, c4, d1, d2, d4);
    out[2][2]  =   det3x3( a1, a2, a4, b1, b2, b4, d1, d2, d4);
    out[3][2]  = - det3x3( a1, a2, a4, b1, b2, b4, c1, c2, c4);
        
    out[0][3]  = - det3x3( b1, b2, b3, c1, c2, c3, d1, d2, d3);
    out[1][3]  =   det3x3( a1, a2, a3, c1, c2, c3, d1, d2, d3);
    out[2][3]  = - det3x3( a1, a2, a3, b1, b2, b3, d1, d2, d3);
    out[3][3]  =   det3x3( a1, a2, a3, b1, b2, b3, c1, c2, c3);
}


/*****************************************************************************
MatInvertFull

A full 4x4 invert, not specialized like the above. About 12% slower...

*****************************************************************************/
int
MatInvertFull(CMatrix out, CMatrix in)
{
    int i, j;
    double det;

    /* calculate the adjoint matrix */

    adjoint( in, out );

    /*  calculate the 4x4 determinent
     *  if the determinent is zero, 
     *  then the inverse matrix is not unique.
     */

    det = det4x4( in );

    if ( fabs( det ) < 1.e-8) {
        fprintf(stderr,"MatInvertFull() No inverse!\n");
        return (FALSE);
    }

    /* scale the adjoint matrix to get the inverse */

    for (i=0; i<4; i++)
        for(j=0; j<4; j++)
	    out[i][j] = out[i][j] / det;

    return (TRUE);
}


/*****************************************************************************
MatInvertAnglePreserving
  
 Computes the inverse of a 3-D angle-preserving matrix. Over 8 times
 faster than MatInvert().
 
 This procedure treats the 4 by 4 angle-preserving matrix as a block
 matrix and calculates the inverse of one submatrix for a significant
 performance improvement over a general procedure that can invert any
 nonsingular matrix:
          --        --          --                   --
          |          | -1       |   -2 T      -2  T   |
          | A      C |          |  s  A     - s  A  C |
    -1    |          |          |                     |
   M   =  |          |     =    |                     |
          | 0      1 |          |    0           1    |
          |          |          |                     |
          --        --          --                   --
 where      M is a 4 by 4 angle-preserving matrix,
            A is the 3 by 3 upper-left submatrix of M,
            C is the 3 by 1 upper-right submatrix of M.
 
 Input:
   in   - 3-D angle-preserving matrix

 Output:
   out  - inverse of 3-D angle-preserving matrix
  
 Returned value:
   TRUE   if input matrix is nonsingular
   FALSE  otherwise

*****************************************************************************/
int
MatInvertAnglePreserving(CMatrix out, CMatrix in)
{
    float  scale;

    /* Calculate the square of the isotropic scale factor */
    scale = in[0][0] * in[0][0] +
            in[0][1] * in[0][1] +
            in[0][2] * in[0][2];

    /* Is the submatrix A singular? */
    if (scale == 0.0) {

        /* CMatrix M has no inverse */
        fprintf(stderr, "MatInvertAnglePreserving() No inverse!\n");
        return (FALSE);
    }

    /* Calculate the inverse of the square of the isotropic scale factor */
    scale = 1.0 / scale;

    /* Transpose and scale the 3 by 3 upper-left submatrix */
    out[0][0] = scale * in[0][0];
    out[1][0] = scale * in[0][1];
    out[2][0] = scale * in[0][2];
    out[0][1] = scale * in[1][0];
    out[1][1] = scale * in[1][1];
    out[2][1] = scale * in[1][2];
    out[0][2] = scale * in[2][0];
    out[1][2] = scale * in[2][1];
    out[2][2] = scale * in[2][2];

    /* Calculate -(transpose(A) / s*s) C */
    out[0][3] = - ( out[0][0] * in[0][3] +
                    out[0][1] * in[1][3] +
                    out[0][2] * in[2][3] );
    out[1][3] = - ( out[1][0] * in[0][3] +
                    out[1][1] * in[1][3] +
                    out[1][2] * in[2][3] );
    out[2][3] = - ( out[2][0] * in[0][3] +
                    out[2][1] * in[1][3] +
                    out[2][2] * in[2][3] );

    /* Fill in last row */
    out[3][0] = out[3][1] = out[3][2] = 0.0;
    out[3][3] = 1.0;

    return (TRUE);
}


/*****************************************************************************
MatZero

Zero out the matrix.

*****************************************************************************/
void
MatZero(CMatrix mat)
{
   int i,j;

   for (i=0; i<4; i++)
     for (j=0; j<4; j++)
        mat[i][j] = 0.0;
}


/*****************************************************************************
MatIdentity

Make an identity matrix.

*****************************************************************************/
void
MatIdentity(CMatrix mat)
{
   mat[0][0] = 1.0;
   mat[0][1] = 0.0;
   mat[0][2] = 0.0;
   mat[0][3] = 0.0;

   mat[1][0] = 0.0;
   mat[1][1] = 1.0;
   mat[1][2] = 0.0;
   mat[1][3] = 0.0;

   mat[2][0] = 0.0;
   mat[2][1] = 0.0;
   mat[2][2] = 1.0;
   mat[2][3] = 0.0;

   mat[3][0] = 0.0;
   mat[3][1] = 0.0;
   mat[3][2] = 0.0;
   mat[3][3] = 1.0;
}


/*****************************************************************************
MatMultiply

This multiplies two matrices, producing a third, which may the same as
either of the first two.

*****************************************************************************/
void
MatMultiply(CMatrix result, CMatrix mat1, CMatrix mat2)
{
   register int i, j;
   CMatrix       tmp_mat;
   float       *tmp;

   if (((tmp = (float*)result) == (float*)mat1) ||
        (tmp == (float*)mat2))
      tmp = (float*)tmp_mat;

   for (i = 0; i < 4; i++) for (j = 0; j < 4; j++)
      *(tmp++) = (mat1[i][0] * mat2[0][j] +
                  mat1[i][1] * mat2[1][j] +
                  mat1[i][2] * mat2[2][j] +
                  mat1[i][3] * mat2[3][j]);

   if (((float*)result == (float*)mat1) ||
       ((float*)result == (float*)mat2))
      MatCopy(result, tmp_mat);
}


/*****************************************************************************
MatSetTranslate

Just set the translate field of the matrix given.

*****************************************************************************/
void
MatSetTranslate(CMatrix mat, CVector pos)
{
   mat[0][3] = pos[0];
   mat[1][3] = pos[1];
   mat[2][3] = pos[2];
}


/*****************************************************************************
MatGetTranslate

Just get the translate field of the matrix given.

*****************************************************************************/
void
MatGetTranslate(CVector pos, CMatrix mat)
{
   pos[0] = mat[0][3];
   pos[1] = mat[1][3];
   pos[2] = mat[2][3];
}


/*****************************************************************************
MatTranslate

Translate the matrix by the vector given.

*****************************************************************************/
void
MatTranslate(CMatrix mat, CVector pos)
{
   mat[0][3] += pos[0];
   mat[1][3] += pos[1];
   mat[2][3] += pos[2];
}


/*****************************************************************************
MatTranslate

Translate the matrix by the x,y,z given.

*****************************************************************************/
void
MatTranslate(CMatrix mat, float tx, float ty, float tz)
{
   mat[0][3] += tx;
   mat[1][3] += ty;
   mat[2][3] += tz;
}


/*****************************************************************************
MatScale

Apply a pre-scale matrix to the matrix given.

*****************************************************************************/
void
MatScale(CMatrix mat, CVector vec)
{
   CMatrix tmp;

   MatIdentity(tmp);
   tmp[0][0] = vec[0];
   tmp[1][1] = vec[1];
   tmp[2][2] = vec[2];

   MatMultiply(mat,mat,tmp);
}


/*****************************************************************************
MatScaleUniform

Apply a uniform pre-scale to the matrix given.

*****************************************************************************/
void
MatScaleUniform(CMatrix mat, float s)
{
   CMatrix tmp;

   MatIdentity(tmp);
   tmp[0][0] = s;
   tmp[1][1] = s;
   tmp[2][2] = s;

   MatMultiply(mat,mat,tmp);
}


/*****************************************************************************
MatRotate

Creates a matrix for a rotation about an axis given by the line from
(0,0,0) to axis, through an angle (in radians). Looking along the axis 
toward the origin, the rotation is counter-clockwise.

*****************************************************************************/
void
MatRotate(CMatrix result, CVector axis, float angle_in_radians)
{
   CVector	base1,  /* 1st unit basis vec, normalized axis          */
                base2,	/* 2nd unit basis vec, perp to axis		*/
	        base3;	/* 3rd unit basis vec, perp to axis & base2	*/
   CMatrix	base_mat,	/* Change-of-basis matrix		*/
		base_mat_trans; /* Inverse of c-o-b matrix		*/
   register int i;

   /* Step 1: extend { axis } to a basis for 3-space: { axis, base2, base3 }
    * which is orthonormal (all three have unit length, and all three are
    * mutually orthogonal). Also should be oriented, i.e. axis cross base2 =
    * base3, rather than -base3.
    *
    * Method: Find a vector linearly independent from axis. For this we
    * either use the y-axis, or, if that is too close to axis, the
    * z-axis. 'Too close' means that the dot product is too near to 1.
    */

   VecCopy(base1,axis);

   VecNormalize(base1);
#if 0
   if (VecNormalize(base1) == FALSE) {
   
      /* If the axis was too short to normalize (i.e., it is almost
       * a zero length vector), just return the identity matrix (no rotation).
       */
      MatIdentity(result);
      return;
   }
#endif

   VecPerp        (base2, base1, TRUE);
   VecCrossProduct(base3, base1, base2);

   /* Set up the change-of-basis matrix, and its inverse */
   MatIdentity(base_mat);
   MatIdentity(base_mat_trans);
   MatIdentity(result);

   for (i = 0; i < 3; i++){
      base_mat_trans[i][0] = base_mat[0][i] = base1[i];
      base_mat_trans[i][1] = base_mat[1][i] = base2[i];
      base_mat_trans[i][2] = base_mat[2][i] = base3[i];
   }

   /* If T(u) = uR, where R is base_mat, then T(x-axis) = axis,
    * T(y-axis) = base2, and T(z-axis) = base3. The inverse of base_mat is
    * its transpose.  OK?
    */

   result[1][1] =   result[2][2] = COS(angle_in_radians);
   result[2][1] = -(result[1][2] = SIN(angle_in_radians));

   MatMultiply(result, base_mat_trans, result);
   MatMultiply(result, result,	       base_mat);
}


/*****************************************************************************
MatRotateAxis

Create a rotation matrix about one of the principal axis.

*****************************************************************************/
void
MatRotateAxis(CMatrix mat, int axis, float radians)
{
   float  cos_theta, sin_theta;

   MatIdentity(mat);

   sin_theta = SIN(radians);
   cos_theta = COS(radians);

   switch (axis) {
     case 0:
         mat[1][1] = cos_theta;
         mat[2][2] = cos_theta;
         mat[2][1] = sin_theta;
         mat[1][2] = -1.* sin_theta;
         break;

     case 1:
         mat[0][0] = cos_theta;
         mat[2][2] = cos_theta;
         mat[2][0] = -1.* sin_theta;
         mat[0][2] = sin_theta;
         break;

     case 2:
         mat[0][0] = cos_theta;
         mat[1][1] = cos_theta;
         mat[1][0] = sin_theta;
         mat[0][1] = -1.* sin_theta;
         break;
   }
}


/*****************************************************************************
MatCreateOrthoRotate

Create an orthogonal rotation matrix.

*****************************************************************************/
void
MatCreateOrthoRotate(CMatrix mat, CVector a, CVector b, CVector c)
{
   register int i;

   MatIdentity(mat);
   for (i = 0; i < 3; i++) {
      mat[0][i] = a[i];
      mat[1][i] = b[i];
      mat[2][i] = c[i];
   }
}


/*****************************************************************************
MatCreateViewMatrix

Create a view matrix defined by an eye point, view reference point
and up vector (terminology as defined in Foley and van Dam).

NOTE: currently used by splinePath class, thus I assume the code works?

*****************************************************************************/
void
MatCreateViewMatrix(CMatrix mat, CVector eye, CVector vrp, CVector vup)
{
   CVector udir,wdir,vdir;
   CMatrix temp,temp2;

   VecSubtract(wdir,eye,vrp);
   MatIdentity(temp);
   temp[0][3] = -eye[0];
   temp[1][3] = -eye[1];
   temp[2][3] = -eye[2];
   VecCrossProduct(udir,wdir,vup);
   VecCrossProduct(vdir,udir,wdir);
   VecNormalize(udir);
   VecNormalize(vdir);
   VecNormalize(wdir);
   VecScale(udir,-1.0,udir);
   MatCreateOrthoRotate(temp2,udir,vdir,wdir);
   MatMultiply(mat,temp2,temp);
}


/*****************************************************************************
MatCreateFrustum

Create a view frustum matrix (not necessary z-axis aligned), ala OpenGL.

*****************************************************************************/
void
MatCreateFrustum(CMatrix mat, 
                 float left,    float right, 
                 float bottom,  float top, 
                 float nearval, float farval)
{
   float x = (2.0*nearval) / (right-left);
   float y = (2.0*nearval) / (top-bottom);
   float a = (right+left) / (right-left);
   float b = (top+bottom) / (top-bottom);
   float c = -(farval+nearval) / ( farval-nearval);
   float d = -(2.0*farval*nearval) / (farval-nearval);  /* error? */

#define M(row,col)  mat[row][col]    /* this lib does transpose of GL style */
   M(0,0) = x;     M(0,1) = 0.0F;  M(0,2) = a;      M(0,3) = 0.0F;
   M(1,0) = 0.0F;  M(1,1) = y;     M(1,2) = b;      M(1,3) = 0.0F;
   M(2,0) = 0.0F;  M(2,1) = 0.0F;  M(2,2) = c;      M(2,3) = d;
   M(3,0) = 0.0F;  M(3,1) = 0.0F;  M(3,2) = -1.0F;  M(3,3) = 0.0F;
#undef M
}


/*****************************************************************************
MatCreatePerspective

Create a perspective transformation matrix, ala OpenGL.

*****************************************************************************/
void
MatCreatePerspective(CMatrix mat, float fovy, float aspect, 
                     float zNear, float zFar)
{
   double xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * M_PI / 360.0 );
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   MatCreateFrustum( mat, xmin, xmax, ymin, ymax, zNear, zFar );
}


/*****************************************************************************
MatPrint

Print out a 4x4 matrix.

*****************************************************************************/
void
MatPrint(char *msge, CMatrix mat)
{
   int i;

   fprintf(stderr,"%s:\n",msge);
   for (i=0; i<4; i++)
      fprintf(stderr,"%1.3f %1.3f %1.3f %1.3f\n",
            mat[i][0],mat[i][1],mat[i][2],mat[i][3]);
}

