#include "cvector.h"


/*****************************************************************************

Computes a unit vector perpendicular to "vec", storing it in "result".
If "is_unit" is TRUE, "vec" is assumed to be of unit length (possibly
allowing a time savings by avoiding normalization of "result").

The method used takes advantage of a technique pointed out by JFH.  As it
is easier to compute a perpendicular vector in R2 than R3, we essentially
do that.  We first determine the component of "vec" with the mimimum
absolute value, setting corresponding component of "result" to 0.
We next set the other two components of "result" by swapping the
remaining (non-mimimum) components of "vec" and negating one of these.

*****************************************************************************/
void
VecPerp(CVector result, CVector vec, int is_unit)
{
   float abs_vec[3];
   int   min;
   int   other1;
   int   other2;

   /* Compute the absolute value of the components of the input vector. */

   abs_vec[0] = ABS(vec[0]);
   abs_vec[1] = ABS(vec[1]);
   abs_vec[2] = ABS(vec[2]);

   /* Determine the component with the minimum absolute value, saving its
      index in "min".  Save the indexes of the other two components in
      "other1" and "other2". */

   if (abs_vec[0] <= abs_vec[1])
      if (abs_vec[0] <= abs_vec[2]) {
         min = 0; other1 = 1; other2 = 2;
      }
      else {
         min = 2; other1 = 0; other2 = 1;
      }
   else
      if (abs_vec[1] <= abs_vec[2]) {
         min = 1; other1 = 0; other2 = 2;
      }
      else {
         min = 2; other1 = 0; other2 = 1;
      }

   /* Generate the perpendicular vector by setting the "min" compenent to
      0, and swapping the "other1" and "other2' components from "vec",
      negating one of these. */

   result[min] = 0.0;
   result[other1] = vec[other2];
   result[other2] = -vec[other1];

   /* If the input vector has unit length and the component with the mimimum
      absolute value is 0, the computed perpendicular vector already has
      unit length.  Otherwise, we must normalize it. */

   if (!is_unit ||
       vec[min] != 0.0)
      VecNormalize(result);
}

