#include "vball.h"

#define BOX1_LEFTX       0       /* left stick x-axis    */
#define BOX1_LEFTY       1       /* left stick y-axis    */
#define BOX1_LEFTZ       2       /* left stick z-axis    */
#define BOX1_RIGHTX      3       /* right stick x-axis   */
#define BOX1_RIGHTY      4       /* right stick y-axis   */
#define BOX1_RIGHTZ      5       /* right stick z-axis   */
#define BOX1_SLIDER      0       /* slider of left joystick box*/

#ifndef NOJOY
/* px joystick magic */
       int px_errno=0;
extern "C" int px_initjoys(void);
extern "C" int px_getjoys(float *, float *);
#endif


/*****************************************************************************
intersectSphere

Intersect the ray defined by 'base' with the sphere (ctr, radius). Return
the point in 'pt'.
 
*****************************************************************************/
int
VBall::intersectSphere(CVector base, CVector ctr, float radius, 
                       CVector dir, CVector pt)
{
   CVector adj,tmp;
   float b,s,t;

   VecSubtract(adj,base,ctr);

   b = VecDotProduct(adj,dir);
   t = SQR(b) - VecLengthSqr(adj) + SQR(radius);
   if (t < 0.0) return (FALSE);
   s = -b - SQRT(t);
   if (s <= 0.0) {
      s = -b + SQRT(t);
      s = ABS(s);
   }

   VecScale(tmp,s,dir);
   VecAdd(pt,base,tmp);

   return (TRUE);
}


/*****************************************************************************
rotateToPoint

Compute the axis and angle about which to rotate the trackball matrix 
based on the point p.

*****************************************************************************/
int
VBall::rotateToPoint(CVector p, CVector axis, float *radians)
{
   CVector base0,base1;
   CVector p0,p1;
   CVector dir0,dir1;
   CVector v0,v1;

   VecSet(base0, ppos[0], ppos[1], 0.0);
   VecSet(dir0,  0.0,     0.0,     1.0);
   VecSet(base1, p[0],    p[1],    0.0);
   VecSet(dir1,  0.0,     0.0,     1.0);

   if (!intersectSphere(base0,ctr,radius,dir0,p0))
      return (FALSE);
   if (!intersectSphere(base1,ctr,radius,dir1,p1))
      return (FALSE);

   VecSubtract(v0,p0,ctr);
   VecSubtract(v1,p1,ctr);

   VecNormalize(v0);
   VecNormalize(v1);

   VecCrossProduct(axis,v0,v1);

   (*radians) = asin((double)VecLength(axis));

   if (VecDotProduct(v0,v1) < 0)
      (*radians) += 3.1415926535897932384/2.0;

   VecCopy(ppos,p);

   return (TRUE);
}


/*****************************************************************************
Constructor

Create a new trackball in a window of size width x height, assuming the model
radius given.

*****************************************************************************/
VBall::VBall(float rad, int in_width, int in_height)
{
   VecSet(ppos,0.0,0.0,0.0);
   VecSet(ctr,0,0,0);
   first      = TRUE;
   radius     = rad;
   width      = in_width;
   height     = in_height;
   walkMode   = FALSE;
   initJoy    = TRUE;
#ifndef NOJOY
   deltaRot   = VBALL_DFLT_DELTAROT;
   deltaTrans = VBALL_DFLT_DELTATRANS;
#endif

   MatIdentity(ctm);
} 


/*****************************************************************************
walkRotate

Use the delta x to perform a y-axis rotate operation.

*****************************************************************************/
void 
VBall::walkRotate(int deltaX, float scale)
{
   CMatrix rotation;

   MatRotateAxis(rotation, 1, deltaX * scale);
   MatMultiply(ctm,rotation,ctm);
}


/*****************************************************************************
rotate

Use the delta x,y to perform a rotate operation.

*****************************************************************************/
void 
VBall::rotate(int x, int y, float scale)
{
   CMatrix ictm,rotation;
   CVector p,axis;
   float  radians;
   float  xstep = 0.5 * (2*radius) / (float)width;
   float  ystep = 0.5 * (2*radius) / (float)height;

   VecSet(p, (x * xstep) - radius,
             (y * ystep) - radius,
             0.0);

   if (first == TRUE) {
      oldx = x;
      oldy = y;
      VecCopy(ppos,p);
      first = FALSE;
      return;
   }

   if (!rotateToPoint(p, axis, &radians))
      radians = 0.0;

   /* --- negate radians here! --- */
   radians *= -scale;
   /* --- negate radians here! --- */

   if (ABS(radians) > FLOAT_EPSILON) {
      MatInvertAnglePreserving(ictm,ctm);
      MatMultVec(axis,ictm,axis);

      MatGetTranslate(p,ctm);
      VecNegate(p,p);
      MatTranslate(ctm,p);
      MatRotate(rotation,axis,radians);
      MatMultiply(ctm,ctm,rotation);
      VecNegate(p,p);
      MatTranslate(ctm,p);
   }

   oldx = x;
   oldy = y;
}


/*****************************************************************************
translate

Use the delta x,y to perform a X,Y translate operation.

*****************************************************************************/
void 
VBall::xytranslate(int x, int y)
{
   CVector p;

   if (first == TRUE) {
      oldx = x;
      oldy = y;
      first = FALSE;
      return;
   }

   VecSet(p, (float)(x - oldx) / (float)(width<<1),
             (float)(y - oldy) / (float)(height<<1),
             0.0);
   VecScale(p, radius, p);
   MatTranslate(ctm,p);

   oldx = x;
   oldy = y;
}


/*****************************************************************************
ztranslate

Use the delta y to perform a Z translate operation.

*****************************************************************************/
void 
VBall::ztranslate(int y)
{
   CVector p;
   float  z,fy = -y;

   z = fy;

   if (first == TRUE) {
      oldz = fy;
      first = FALSE;
      return;
   }

   VecSet(p, 0.0, 0.0, (z - oldz) / width);
   VecScale(p, radius, p);

   MatTranslate(ctm,p);

   oldz = fy;
}


/*****************************************************************************
scale

Use the MAX(delta x,delta y) to perform a uniform scale operation.

*****************************************************************************/
void 
VBall::scale(int x, int y)
{
   float  amount,sfact;
   CVector scale;

   if (first == TRUE) {
      oldx = x;
      oldy = y;
      first = FALSE;
      return;
   }

   if (ABS(x - oldx) > ABS(y - oldy))
      amount = x - oldx;
   else
      amount = y - oldy;
   sfact = 1.0 + (amount/(float)((width+height) >> 1));
   VecSet(scale,sfact,sfact,sfact);
   MatScale(ctm,scale);

   oldx = x;
   oldy = y;
}


/*****************************************************************************
xscale

Use the delta x to perform a uniform scale operation.

*****************************************************************************/
void 
VBall::xscale(int x)
{
   float  amount,sfact;
   CVector scale;

   if (first == TRUE) {
      oldx = x;
      first = FALSE;
      return;
   }

   amount = x - oldx;
   sfact = 1.0 + amount/(float)width;
   VecSet(scale,sfact,1.0,1.0);
   MatScale(ctm,scale);

   oldx = x;
}


/*****************************************************************************
yscale

Use the delta y to perform a uniform scale operation.

*****************************************************************************/
void 
VBall::yscale(int y)
{
   float  amount,sfact;
   CVector scale;

   if (first == TRUE) {
      oldy = y;
      first = FALSE;
      return;
   }

   amount = y - oldy;
   sfact = 1.0 + amount/(float)height;
   VecSet(scale,1.0,sfact,1.0);
   MatScale(ctm,scale);

   oldy = y;
}


/*****************************************************************************
vballSpin

Spin the view about the axis given for the radians given. The walk flag
indicates whether to spin about the origin or about the trackball center.

*****************************************************************************/
void 
VBall::spin(int axis, float radians)
{
   CMatrix rotation;
   CVector p;

   if (walkMode && radians != 0.0) {

      MatRotateAxis(rotation,axis,radians);
      MatMultiply(ctm,rotation,ctm);

   } else if (!walkMode && radians != 0.0) {

      MatGetTranslate(p,ctm);
      VecNegate(p,p);
      MatTranslate(ctm,p);
      MatRotateAxis(rotation,axis,radians);
      MatMultiply(ctm,ctm,rotation);
      VecNegate(p,p);
      MatTranslate(ctm,p);
   }
}


/*****************************************************************************
initJoysticks

Initialize joysticks. If initialize fails, set initJoy to -1.

*****************************************************************************/
void
VBall::initJoysticks()
{
#ifdef NOJOY
   initJoy = -1;
#else
   if (px_initjoys() < 0) {
      ERROR((STDERR,"VBall::initJoystick() Could not initialize "
                    "joysticks()\n"));
      initJoy = -1;
   } else {
      initJoy = FALSE;
   }
#endif
}


/*****************************************************************************
makeMat

Create a matrix to append to the current view matrix, using the current
joystick state.

Parameters:

mat        matrix to be created                 
rx, ry, rz rotation vector (degrees)            
tx, ty, tz translation vector                   

*****************************************************************************/
void
VBall::makeMat(CMatrix mat, float rx, float ry, float rz, 
                           float tx, float ty, float tz)
{
    float  theta;     /* non-y rotation angle  */
    CMatrix R0;        /* non-y rotation matrix */
    CMatrix R1;        /* y rotation matrix     */
    CMatrix transMat;  /* translation matrix    */
    CVector axis;      /* generic rotation axis */

    MatIdentity(mat);

    rx *= deltaRot;
    ry *= deltaRot;
    rz *= deltaRot;
    tx *= deltaTrans;
    ty *= deltaTrans;
    tz *= deltaTrans;

    /* rotate does not work right, TBS problem */
#if 0
    /* non-y rotation */
    theta = SQRT(rx*rx + rz*rz);
#else
    theta = 0.0;
#endif

    if (theta == 0) {
        MatIdentity(R0);
    } else {

        VecSet(axis, rx/theta, 0.0, rz/theta);
        VecNormalize(axis);
        MatRotate(R0, axis, theta);
    }

    VecSet(axis,0,1,0);
    MatRotate(R1, axis, ry);

    /* composite */
    MatMultiply(mat, R1, R0);

    /* Add the translation factors */
    MatIdentity(transMat);
    MatTranslate(transMat, tx, ty, tz);

    MatMultiply(mat, mat, transMat);
}


/*****************************************************************************
updateJoysticks

If joysticks can be initialized, this will append to the current modelview
matrix the transformation given by the walkthrough controls of the joystick.

*****************************************************************************/
int
VBall::updateJoysticks()
{
#ifdef NOJOY
   return (FALSE);
#else
   CMatrix postMat;
   int    joysMoved;
   float  joys[12], slider[2];

   /* init joysticks */
   if (initJoy == TRUE) initJoysticks();

   /* if failed, just return */
   if (initJoy < 0) return (FALSE);

   /* get joystick state */ 
   joysMoved = px_getjoys(slider, joys);

   /* if changed, construct matrix */
   if (joysMoved) {
      /* get deltaTrans from slider */
      deltaTrans = 1.0 * (slider[BOX1_SLIDER]+0.5) *
                         (slider[BOX1_SLIDER]+0.5);
      makeMat(postMat, 
              -joys[BOX1_LEFTY], -joys[BOX1_RIGHTZ], 0.0,
              -joys[BOX1_RIGHTX], -joys[BOX1_LEFTZ],  joys[BOX1_RIGHTY]);

      /* append to modelview mat */
      MatMultiply(ctm, postMat, ctm); 
   }

   return (joysMoved);
#endif
}


/*****************************************************************************
setLevelEye

Force the eye to be viewing parallel to the XZ plane (i.e. level).

NOTE: only works if less < PI/2 from level plane.

*****************************************************************************/
void
VBall::setLevelEye()
{
   CVector eye,level;
   CMatrix rotMat;
   float  theta;

   /* compute angle between -Z and eye */
   VecSet(eye,0,0,-1);
   MatMultVec(eye, ctm, eye);
   VecSet(level, eye[0], 0, eye[2]);
   VecNormalize(eye);
   VecNormalize(level);
   theta = ACOS(VecDotProduct(eye,level));

   /* compute rotation matrix */
   if (eye[1] > 0) theta *= -1;
   MatRotateAxis(rotMat, 0, theta);

   /* append to modelview mat */
   MatMultiply(ctm, rotMat, ctm); 
}
