#ifndef VMARRAY_ALREADY_INCLUDED
#define VMARRAY_ALREADY_INCLUDED

#include <common.h>


/*****************************************************************************

  This template class implements a dynamic array. The array increases
  and decreases in size automatically, can be accessed using the [] operator,
  copied, and includes operations like prepend(), append(), insert(), find(), 
  remove(), and replace().

*****************************************************************************/
template <class VMArrayEntry>
class VMArray {
   private:
      int maxSize;
      int currSize;
      VMArrayEntry *array;

      //////
      // This method handles increasing the size of the array
      //////
      void increase()
      {
         if (maxSize == 0) {
            // if first append ever, create initial array
            currSize = 0;
            maxSize = 2;
            array = new VMArrayEntry[maxSize];
         } else if (currSize == maxSize) {
            // if needed, expand array size
            VMArrayEntry *oldArray = array;
            maxSize *= 2;
            array = new VMArrayEntry[maxSize];
            if (oldArray) {
               for (int n=0; n<currSize; n++)
                  array[n] = oldArray[n];
               delete [] oldArray;
            }
         }
      }


      //////
      // This method handles decreasing the size of the array
      //////
      void decrease()
      {
         if (currSize == 0) {
            // if appropriate, clear out the array
            maxSize = 0;
            if (array) delete [] array;
            array = NULL;
         } else if (currSize < (maxSize >> 1)) {
            // if necessary, reduce array size
            VMArrayEntry *oldArray = array;
            maxSize = maxSize >> 1;
            array = new VMArrayEntry[maxSize];
            if (oldArray) {
               for (int n=0; n<currSize; n++)
                  array[n] = oldArray[n];
               delete [] oldArray;
            }
         }
      }


   public:

      //////
      // Constructors and destructor
      //////
      VMArray()
      {
         maxSize = currSize = 0;
         array = NULL;
      }
      VMArray(int size)
      {
         maxSize = currSize = size;
         array = new VMArrayEntry[maxSize];
      }
      VMArray(VMArray &ref)
      {
         maxSize = ref.maxSize;
         currSize = ref.currSize;
         array = new VMArrayEntry[maxSize];
         for (int n=0; n<currSize; n++)
            array[n] = ref.array[n];
      }
      ~VMArray()
      {
         if (array) delete [] array;
         array = NULL;
      }


      //////
      // Methods for added/replacing/finding/removing data
      //////
      int append(VMArrayEntry &data)
      {
         increase();
         array[currSize] = data;
         currSize++;

         return (currSize-1);
      }
      int prepend(VMArrayEntry &data)
      {
         increase();
         for (int k=currSize-1; k>=0; k--)
            array[k+1] = array[k];
         array[0] = data;
         currSize++;

         return (0);
      }
      void insert(int index, VMArrayEntry &data)
      {
         IFDEBUG(index < 0 || index >= currSize) {
            fprintf(stderr, "VMArray: array index (%i) out of bounds\n", index);
            return;
         }
         increase();
         for (int k=currSize-1; k>=index; k--)
            array[k+1] = array[k];
         array[index] = data;
         currSize++;
      }
      int replace(int index, VMArrayEntry &data)
      {
         IFDEBUG(index < 0 || index >= currSize) {
            fprintf(stderr, "VMArray: array index (%i) out of bounds\n", index);
            return (FALSE);
         }
         array[index] = data;
         return (TRUE);
      }
      int remove(const int index)
      {
         IFDEBUG(index < 0 || index >= currSize) {
            fprintf(stderr, "VMArray: array index (%i) out of bounds\n", index);
            return (FALSE);
         }
         for (int k=index; k<(currSize-1); k++)
            array[k] = array[k+1];
         currSize--;
         decrease();
         return (TRUE);
      }
      int remove(VMArrayEntry &entry)
      {
         int ret = FALSE;
         for (int k=0; k<currSize; k++) {
            if (entry == array[k]) {
               ret = remove(k);
               break;
            }
         }
         return (ret);
      }
      int find(VMArrayEntry &entry)
      {
         for (int k=0; k<currSize; k++) {
            if (array[k] == entry) 
               return (k);
         }
         return (-1);
      }


      //////
      // Methods for accessing data
      //////
      inline VMArrayEntry &get(int index)
      {
         IFDEBUG(index < 0 || index >= currSize) {
            fprintf(stderr, "VMArray: array index (%i) out of bounds\n", index);
            return array[0];
         }

         return (array[index]);
      }


      //////
      // Operator overloading
      //////
      inline VMArray &operator=(VMArray &ref)
      {
         maxSize = ref.maxSize;
         currSize = ref.currSize;
         if (array) delete [] array;
         array = new VMArrayEntry[maxSize];
         for (int n=0; n<currSize; n++)
            array[n] = ref.array[n];
         return (*this);
      }
      inline VMArrayEntry &operator[](int index)
      {
         IFDEBUG(index < 0 || index >= currSize) {
            fprintf(stderr, "VMArray: array index (%i) out of bounds\n", index);
            return array[0];
         }

         return (array[index]);
      }
      inline VMArray &operator+=(VMArrayEntry& data) 
      {
         append(data);
         return *this;
      }
      inline VMArray &operator-=(VMArrayEntry& data) 
      {
         remove(data);
         return *this;
      }



      //////
      // Misc methods
      //////
      int size() 
      { 
         return (currSize); 
      }
      void empty() 
      {
         if (array) delete [] array;
         array = NULL;
         maxSize = currSize = 0;
      }
      void fit() 
      {
         if (currSize != maxSize) {
            VMArrayEntry *oldArray = array;
            maxSize = currSize;
            array = new VMArrayEntry[maxSize];
            if (oldArray) {
               for (int n=0; n<currSize; n++)
                  array[n] = oldArray[n];
               delete [] oldArray;
            }
         }
      }
      void resize(int size)
      {
         if (currSize != size) {
            VMArrayEntry *oldArray = array;
            maxSize = size;
            array = new VMArrayEntry[maxSize];
            if (oldArray) {
               for (int n=0; n<maxSize && n<currSize; n++)
                  array[n] = oldArray[n];
               delete [] oldArray;
            }
            currSize = size;
         }
      }
      void clear()
      {
         BZERO(array, maxSize * sizeof(VMArrayEntry));
      }

};


#endif
