#ifndef VM2MATRIX_ALREADY_INCLUDED
#define VM2MATRIX_ALREADY_INCLUDED

#include <vm2vector.h>
#include <vm2point.h>


#define VM2MATRIX_ROTATE_X_AXIS        0
#define VM2MATRIX_ROTATE_Y_AXIS        1

#define VM2MATRIX_RAD_TO_DEG(R)        (((R)/M_PI)*180.0)
#define VM2MATRIX_DEG_TO_RAD(D)        (((D)/180.0)*M_PI)


/*****************************************************************************

This class implements a standard 2x2 matrix. Operations are assumed to be 
performed with nonhomogenous 2D coordinates (using VM2Vector/VM2Point). 
Supported operations are: scale, rotate, invert.

*****************************************************************************/

class VM2Matrix {
      float mat[2][2];

   public:

      /* 
       * Constructors 
       */
      VM2Matrix() 
      { 
         mat[0][0] = 1.0; mat[0][1] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0;
      }

      VM2Matrix(float a, float b,
                float c, float d)
      {
         mat[0][0] = a; mat[0][1] = b;
         mat[1][0] = c; mat[1][1] = d;
      }

      VM2Matrix(float m[2][2]) 
      {
         BCOPY(m, mat, sizeof(float)*4);
      }

      VM2Matrix(const VM2Matrix &m) 
      { 
         BCOPY(m.mat, mat, sizeof(float)*4);
      }


      /*
       * Matrix operations and utilities
       */
      void identity() 
      {
         mat[0][0] = 1.0; mat[0][1] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0;
      }

      void zero() 
      {
         BZERO(mat, sizeof(float)*4);
      }

      void transpose() 
      {
        int   r,c;
        float tmp;

        for (r = 0; r < 2; r++) {
           for (c = r; c < 2; c++) {
              if (r != c) {
                 SWAP(mat[r][c], mat[c][r], tmp);
              }
           }
        }
      }

      /* no translate operation */

      /* scale operation */
      void scale(float x, float y)
      {
         VM2Matrix mat;

         mat.mat[0][0] = x;
         mat.mat[1][1] = y;

         *this = (*this) * mat;
      }
      void scale(const VM2Point  &s) { scale(s.x(),s.y()); }
      void scale(const VM2Vector &s) { scale(s.x(),s.y()); }
      void scale(float s)            { scale(s, s); }

      /* rotation operation */
      void rotate(float radians);

      /* inversion operations */
      int   invert();
      float determinant()
      {
         return (mat[0][0]*mat[1][1] - mat[0][1]*mat[1][0]);
      }

      /* others */
      void set(int r, int c, float val) 
      { 
         IFDEBUG(r < 0 || r >= 2 || c < 0 || c >= 2) {
            fprintf(stderr, "VM2Matrix: index (%i,%i) out of range\n", r, c);
            return;
         }
         mat[r][c] = val; 
      }

      void print(FILE *fp)
      {
         int i,j;

         for (i=0; i<2; i++) {
            for (j=0; j<2; j++)
               fprintf(fp,"%1.4f ", mat[i][j]);
            fprintf(fp,"\n");
         }
         fprintf(fp,"\n");
      }


      /*
       * Operator overloading
       */
      inline float operator()(int r, int c) const
      {
         IFDEBUG(r < 0 || r >= 2 || c < 0 || c >= 2) {
            fprintf(stderr, "VM2Matrix: index (%i,%i) out of range\n", r, c);
            return 0;
         }
         return (mat[r][c]);
      }

      inline VM2Matrix &operator=(const VM2Matrix &m) 
      {
         BCOPY(m.mat, mat, sizeof(float)*4);
         return *this;
      }

      inline VM2Matrix &operator+=(const VM2Matrix &m) 
      {
         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             mat[i][j] += m.mat[i][j];

         return *this;
      }

      inline VM2Matrix &operator-=(const VM2Matrix &m) 
      {
         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             mat[i][j] -= m.mat[i][j];

         return *this;
      }

      inline VM2Matrix &operator*=(const VM2Matrix &m) 
      {
         register float res[2][2];

         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             res[i][j] = mat[i][0] * m.mat[0][j] +
                         mat[i][1] * m.mat[1][j];
         BCOPY(res, mat, sizeof(float)*4);
         return *this;
      }

      inline friend VM2Matrix operator+(const VM2Matrix &a, const VM2Matrix &b) 
      {
         register float res[2][2];

         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             res[i][j] = a.mat[i][j] + b.mat[i][j];
         return (VM2Matrix(res));
      }

      inline friend VM2Matrix operator-(const VM2Matrix &a, const VM2Matrix &b) 
      {
         register float res[2][2];

         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             res[i][j] = a.mat[i][j] - b.mat[i][j];
         return (VM2Matrix(res));
      }

      inline friend VM2Matrix operator*(const VM2Matrix &a, const VM2Matrix &b)
      {
         register float res[2][2];

         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             res[i][j] = a.mat[i][0] * b.mat[0][j] +
                         a.mat[i][1] * b.mat[1][j];
         return (VM2Matrix(res));
      }

      inline friend VM2Matrix operator*(const VM2Matrix &a, float b)
      {
         register float res[2][2];

         for (int i=0; i<2; i++)
          for (int j=0; j<2; j++)
             res[i][j] = a.mat[i][j] * b;
         return VM2Matrix(res);
      }

      inline int operator==(const VM2Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; ret && i<2; i++)
            for (int j=0; ret && j<2; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = FALSE;
         return (ret);
      }

      inline int operator!=(const VM2Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; !ret && i<2; i++)
            for (int j=0; !ret && j<2; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = TRUE;
         return (ret);
      }


      /*
       * Destructors
       */
      ~VM2Matrix() {};
};


#endif
