#include "vm4matrix.h"


/*****************************************************************************

Rotate the matrix about either the x, y or z axis by the given
amount of radians.

*****************************************************************************/
void 
VM4Matrix::rotate(int axis, float radians)
{
   float sinTheta = SIN(radians);
   float cosTheta = COS(radians);
   VM4Matrix m;

   switch (axis) {
      case VM4MATRIX_ROTATE_X_AXIS:
         m.mat[1][1] = cosTheta;
         m.mat[2][2] = cosTheta;
         m.mat[2][1] = sinTheta;
         m.mat[1][2] = -1.* sinTheta;
         break;
   
     case VM4MATRIX_ROTATE_Y_AXIS:
         m.mat[0][0] = cosTheta;
         m.mat[2][2] = cosTheta;
         m.mat[2][0] = -1.* sinTheta;
         m.mat[0][2] = sinTheta;
         break;

     case VM4MATRIX_ROTATE_Z_AXIS:
         m.mat[0][0] = cosTheta;
         m.mat[1][1] = cosTheta;
         m.mat[1][0] = sinTheta;
         m.mat[0][1] = -1.* sinTheta;
         break;
   }
   (*this) = (*this) * m;
}


/*****************************************************************************

Sets up a matrix for a rotation about an axis given by the line from
(0,0,0) through (ax,ay,az), about a rotation of 'radians'. Looking along 
the axis toward the origin, the rotation is counter-clockwise.

*****************************************************************************/
void
VM4Matrix::rotate(const VM3Vector &axis, float radians)
{
   VM3Vector b1,   /* 1st unit basis vec, normalized axis */
             b2,   /* 2nd unit basis vec, perp to axis */
             b3;   /* 3rd unit basis vec, perp to axis & b2 */
   VM4Matrix base_mat,/* Change-of-basis matrix */
             base_mat_trans, /* Inverse of c-o-b matrix */
             result_mat;     /* new rotation matrix */
   int i;

   /* Step 1: extend { axis } to a basis for 3-space: { axis, b2, b3 }
    * which is orthonormal (all three have unit length, and all three are
    * mutually orthogonal). Also should be oriented, i.e. axis cross b2 =
    * b3, rather than -b3.
    *
    * Method: Find a vector linearly independent from axis. For this we
    * either use the y-axis, or, if that is too close to axis, the
    * z-axis. 'Too close' means that the dot product is too near to 1.
    *
    * If axis size is too small, might cause problems.
    */ 

   if (axis.x() == 0.0 && axis.y() == 0.0 && axis.z() == 0.0) {
      fprintf(stderr,"VM4Matrix: cannot rotate about zero vector!\n");
      return;
   }

   /*
    * find orthogonal vectors 
    */
   b1 = axis;
   b1.normalize();
   b2 = b1.perpendicular();
   b2.normalize();
   b3 = b1.cross(b2);
   /* by definition b3 will already be normalized */

   /* Set up the change-of-basis matrix, and its inverse */
   for (i = 0; i < 3; i++) {
      base_mat_trans.mat[i][0] = base_mat.mat[0][i] = b1(i);
      base_mat_trans.mat[i][1] = base_mat.mat[1][i] = b2(i);
      base_mat_trans.mat[i][2] = base_mat.mat[2][i] = b3(i);
   }

   /* If T(u) = uR, where R is base_mat, then T(x-axis) = axis,
    * T(y-axis) = b2, and T(z-axis) = b3. The inverse of base_mat is
    * its transpose.  OK?
    */

   result_mat.mat[2][1] = -(result_mat.mat[1][2] = (float)sin(radians));
   result_mat.mat[1][1] =   result_mat.mat[2][2] = (float)cos(radians);

   result_mat = base_mat_trans * result_mat;
   result_mat = result_mat * base_mat;
   (*this)    = (*this) * result_mat;
}



/*****************************************************************************

Below is the method invert() and its auxiliary routines.

*****************************************************************************/

#define SMALL 1e-8                /* Small enough to be considered zero */
#define MYSWAP(A,B,T)                T=B,B=A,A=T

/*
 * Shuffles rows in inverse of 3x3.  See comment in VM4Matrix_inv3_second_col().
 */
static void
VM4Matrix_inv3_swap(float inv[3][3], int row0, int row1, int row2)
{
   register int i, tempi;
   float        temp;

#define MYSWAP_ROWS(a, b)   for (i = 0; i < 3; i++) MYSWAP(inv[a][i], inv[b][i], temp); MYSWAP(a, b, tempi)

   if (row0 != 0){
      if (row1 == 0) {
         MYSWAP_ROWS(row0, row1);
      }
      else {
         MYSWAP_ROWS(row0, row2);
      }
   }

   if (row1 != 1) {
      MYSWAP_ROWS(row1, row2);
   }
}

/*
 * Does Gaussian elimination on second column.
 */

static int
VM4Matrix_inv3_second_col(float source[3][3], float inv[3][3], int row0)
{
   register int row1, row2, i1, i2, i;
   float        temp;
   float        a, b;

   /* Find which row to use */
   if           (row0 == 0)        i1 = 1, i2 = 2;
   else if (row0 == 1)        i1 = 0, i2 = 2;
   else                 i1 = 0, i2 = 1;

   /* Find which is larger in abs. val.:the entry in [i1][1] or [i2][1]      */
   /* and use that value for pivoting.                                       */

   a = source[i1][1]; if (a < 0) a = -a;
   b = source[i2][1]; if (b < 0) b = -b;
   if (a > b)         row1 = i1;
   else                row1 = i2;
   row2 = (row1 == i1 ? i2 : i1);

   /* Scale row1 in source */
   if ((source[row1][1] < SMALL) && (source[row1][1] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row1][1];
   source[row1][1]  = 1.0;
   source[row1][2] *= temp;        /* source[row1][0] is zero already */

   /* Scale row1 in inv */
   inv[row1][row1]  = temp;        /* it used to be a 1.0 */
   inv[row1][row0] *= temp;

   /* Clear column one, source, and make corresponding changes in inv */

   for (i = 0; i < 3; i++) if (i != row1) {  /* for i = all rows but row1 */
   temp = -source[i][1];
      source[i][1]  = 0.0;
      source[i][2] += temp * source[row1][2];

      inv[i][row1]  = temp * inv[row1][row1];
      inv[i][row0] += temp * inv[row1][row0];
   }

   /* Scale row2 in source */
   if ((source[row2][2] < SMALL) && (source[row2][2] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row2][2];
   source[row2][2] = 1.0;        /* source[row2][*] is zero already */

   /* Scale row2 in inv */
   inv[row2][row2]  = temp;        /* it used to be a 1.0 */
   inv[row2][row0] *= temp;
   inv[row2][row1] *= temp;

   /* Clear column one, source, and make corresponding changes in inv */
   for (i = 0; i < 3; i++) if (i != row2) {  /* for i = all rows but row2 */
   temp = -source[i][2];
      source[i][2] = 0.0;
      inv[i][row0] += temp * inv[row2][row0];
      inv[i][row1] += temp * inv[row2][row1];
      inv[i][row2] += temp * inv[row2][row2];
   }

   /*
    * Now all is done except that the inverse needs to have its rows shuffled.
    * row0 needs to be moved to inv[0][*], row1 to inv[1][*], etc.
    *
    * We *didn't* do the swapping before the elimination so that we could more
    * easily keep track of what ops are needed to be done in the inverse.
    */
   VM4Matrix_inv3_swap(inv, row0, row1, row2);

   return(TRUE);
}

/*
 * Fast inversion routine for 3 x 3 matrices.        - Written by jfh.
 *
 * This takes 30 multiplies/divides, as opposed to 39 for Cramer's Rule.
 * The algorithm consists of performing fast gaussian elimination, by never
 * doing any operations where the result is guaranteed to be zero, or where
 * one operand is guaranteed to be zero. This is done at the cost of clarity,
 * alas.
 *
 * Returns 1 if the inverse was successful, 0 if it failed.
 */

static int
VM4Matrix_invert3(float source[3][3], float inv[3][3])
{
   register int i, row0;
   float        temp;
   float        a, b, c;

   inv[0][0] = inv[1][1] = inv[2][2] = 1.0;
   inv[0][1] = inv[0][2] = inv[1][0] = inv[1][2] = inv[2][0] = inv[2][1] = 0.0;

   /* attempt to find the largest entry in first column to use as pivot */
   a = source[0][0]; if (a < 0) a = -a;
   b = source[1][0]; if (b < 0) b = -b;
   c = source[2][0]; if (c < 0) c = -c;

   if (a > b) {
      if (a > c) row0 = 0;
      else row0 = 2;
   }
   else { 
      if (b > c) row0 = 1;
      else row0 = 2;
   }

   /* Scale row0 of source */
   if ((source[row0][0] < SMALL) && (source[row0][0] > -SMALL)) return(FALSE);
   temp = 1.0 / source[row0][0];
   source[row0][0]  = 1.0;
   source[row0][1] *= temp;
   source[row0][2] *= temp;

   /* Scale row0 of inverse */
   inv[row0][row0] = temp;        /* other entries are zero -- no effort */

   /* Clear column zero of source, and make corresponding changes in inverse */

   for (i = 0; i < 3; i++) if (i != row0) { /* for i = all rows but row0 */
      temp = -source[i][0];
      source[i][0]  = 0.0;
      source[i][1] += temp * source[row0][1];
      source[i][2] += temp * source[row0][2];
      inv[i][row0]  = temp * inv[row0][row0];
   }

   /*
    * We've now done gaussian elimination so that the source and
    * inverse look like this:
    *
    *        1 * *                * 0 0
    *        0 * *                * 1 0
    *        0 * *                * 0 1
    *
    * We now proceed to do elimination on the second column.
    */
   if (! VM4Matrix_inv3_second_col(source, inv, row0)) return(FALSE);

   return(TRUE);
}


/*
 * Finds a new pivot for a non-simple 4x4.  See comments in VM4Matrixinvert().
 */

static int
VM4Matrix_inv4_pivot(float src[4][4], float r[4], float *s, int *swap)
{
   register int i, j;
   float        temp, max;

   *swap = -1;

   if (ABS(src[3][3]) < SMALL) {

      /* Look for a different pivot element: one with largest abs value */
      max = 0.0;

      for (i = 0; i < 4; i++) {
         if         (src[i][3] >  max) max =  src[*swap = i][3];
         else if (src[i][3] < -max) max = -src[*swap = i][3];
      }

      /* No pivot element available ! */
      if (*swap < 0) return(FALSE);

      else for (j = 0; j < 4; j++) MYSWAP(src[*swap][j], src[3][j], temp);
   }

   r[0] = -src[0][3];
   r[1] = -src[1][3];
   r[2] = -src[2][3];

   *s = 1.0 / src[3][3];

   src[0][3] = src[1][3] = src[2][3] = 0.0;
   src[3][3]                             = 1.0;

   src[3][0] *= (*s);
   src[3][1] *= (*s);
   src[3][2] *= (*s);

   for (i = 0; i < 3; i++) {
      src[0][i] += r[0] * src[3][i];
      src[1][i] += r[1] * src[3][i];
      src[2][i] += r[2] * src[3][i];
   }

   return(TRUE);
}


/*
 * This returns the inverse of the given matrix.  The result matrix
 * may be the same as the one to invert.
 *
 * Fast inversion routine for 4 x 4 matrices, written by jfh.
 *
 * Returns 1 if the inverse was successful, 0 if it failed.
 *
 * This routine has been specially tweaked to notice the following:
 * If the matrix has the form
 *          * * * 0
 *          * * * 0
 *          * * * 0
 *          * * * 1
 *
 * (as do many matrices in graphics), then we compute the inverse of
 * the upper left 3x3 matrix and use this to find the general inverse.
 *
 *   In the event that the right column is not 0-0-0-1, we do gaussian
 * elimination to make it so, then use the 3x3 inverse, and then do
 * our gaussian elimination.
 */

int
VM4Matrix::invert()
{
   VM4Matrix            inv;
   register int         i, j, simple;
   float                m[3][3], inv3[3][3], s, temp;
   float                r[4], t[4];
   int                  swap;

   /* INVERT CODE IS ROW MAJOR, SO TRANSPOSE NOW, AND AGAIN AT THE END */
   this->transpose(); 
   /* INVERT CODE IS ROW MAJOR, SO TRANSPOSE NOW, AND AGAIN AT THE END */

   /* If last column is not (0,0,0,1), use special code */
   simple = (mat[0][3] == 0.0 && mat[1][3] == 0.0 &&
             mat[2][3] == 0.0 && mat[3][3] == 1.0);

   if (! simple && ! VM4Matrix_inv4_pivot(mat, r, &s, &swap)) return(FALSE);

   t[0] = mat[3][0];
   t[1] = mat[3][1];
   t[2] = mat[3][2];

   /* Copy upper-left 3x3 matrix */
   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) m[i][j] = mat[i][j];

   if (! VM4Matrix_invert3(m, inv3)) return(FALSE);

   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) inv.mat[i][j] = inv3[i][j];

   for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
      inv.mat[3][i] -= t[j] * inv3[j][i];

   if (! simple) {

      /* We still have to undo our gaussian elimination from earlier on */
      /* add r0 * first col to last col */
      /* add r1 * 2nd        col to last col */
      /* add r2 * 3rd        col to last col */

      for (i = 0; i < 4; i++) {
         inv.mat[i][3] += r[0] * inv.mat[i][0] + 
                          r[1] * inv.mat[i][1] + 
                          r[2] * inv.mat[i][2];
         inv.mat[i][3] *= s;
      }

      if (swap >= 0)
         for (i = 0; i < 4; i++) MYSWAP(inv.mat[i][swap], inv.mat[i][3], temp);
   }

   *this = inv;

   /* INVERT CODE IS ROW MAJOR, SO RE-TRANSPOSE NOW */
   this->transpose(); 
   /* INVERT CODE IS ROW MAJOR, SO RE-TRANSPOSE NOW */

   return(TRUE);
}


/*****************************************************************************
MatInvertAnglePreserving
  
 Computes the inverse of a 3-D angle-preserving matrix. Over 8 times
 faster than invert().
 
 This procedure treats the 4 by 4 angle-preserving matrix as a block
 matrix and calculates the inverse of one submatrix for a significant
 performance improvement over a general procedure that can invert any
 nonsingular matrix:
          --        --          --                   --
          |          | -1       |   -2 T      -2  T   |
          | A      C |          |  s  A     - s  A  C |
    -1    |          |          |                     |
   M   =  |          |     =    |                     |
          | 0      1 |          |    0           1    |
          |          |          |                     |
          --        --          --                   --
 where      M is a 4 by 4 angle-preserving matrix,
            A is the 3 by 3 upper-left submatrix of M,
            C is the 3 by 1 upper-right submatrix of M.
 
 Input:
   in   - 3-D angle-preserving matrix

 Output:
   out  - inverse of 3-D angle-preserving matrix
  
 Returned value:
   TRUE   if input matrix is nonsingular
   FALSE  otherwise

*****************************************************************************/
int
VM4Matrix::invertAnglePreserving()
{
    VM4Matrix out;
    float  scale;

    /* Calculate the square of the isotropic scale factor */
    scale = mat[0][0] * mat[0][0] +
            mat[0][1] * mat[0][1] +
            mat[0][2] * mat[0][2];

    /* Is the submatrix A singular? */
    if (scale == 0.0) {

        /* VM4Matrix M has no inverse */
        fprintf(stderr, "VM4Matrix::invert_angle_preserving() No inverse!\n");
        return (FALSE);
    }

    /* Calculate the inverse of the square of the isotropic scale factor */
    scale = 1.0 / scale;

    /* Transpose and scale the 3 by 3 upper-left submatrix */
    out.mat[0][0] = scale * mat[0][0];
    out.mat[1][0] = scale * mat[0][1];
    out.mat[2][0] = scale * mat[0][2];
    out.mat[0][1] = scale * mat[1][0];
    out.mat[1][1] = scale * mat[1][1];
    out.mat[2][1] = scale * mat[1][2];
    out.mat[0][2] = scale * mat[2][0];
    out.mat[1][2] = scale * mat[2][1];
    out.mat[2][2] = scale * mat[2][2];

    /* Calculate -(transpose(A) / s*s) C */
    out.mat[0][3] = - ( out.mat[0][0] * mat[0][3] +
                        out.mat[0][1] * mat[1][3] +
                        out.mat[0][2] * mat[2][3] );
    out.mat[1][3] = - ( out.mat[1][0] * mat[0][3] +
                        out.mat[1][1] * mat[1][3] +
                        out.mat[1][2] * mat[2][3] );
    out.mat[2][3] = - ( out.mat[2][0] * mat[0][3] +
                        out.mat[2][1] * mat[1][3] +
                        out.mat[2][2] * mat[2][3] );

    /* Fill in last row */
    out.mat[3][0] = out.mat[3][1] = out.mat[3][2] = 0.0;
    out.mat[3][3] = 1.0;

    /* change self */
    *this = out;

    return (TRUE);
}




/*****************************************************************************

  This method assumes the 4x4 matrix is a transformation matrix that consists
  of a rotation followed by a translation.

  This method is useful, for instance, to extract the camera position from a
  standard OpenGL model transformation matrix (note: this code, as all vmmatrix
  code, assumes column-major vectors so it is really the transpose of the
  OpenGL matrices).

NOTE: note sure what happens if the matrix encodes a scale as well. Maybe it
works and just need to "unscale" the return point -- but don't have what to 
unscale it by.

*****************************************************************************/
VM3Point
VM4Matrix::cameraPosition()
{
   // make sure bottom row is 0,0,0,1
   if (mat[3][0] != 0 || mat[3][1] != 0 || mat[3][2] != 0 || mat[3][3] != 1) {
      return VM3Point(0,0,0);
   }

	// by default OpenGL goes from world space to canonical space, so to find 
	// the position of the camera origin (0,0,0) in world space must invert 
	// the matrix
	VM4Matrix iM = *this;
	iM.invertAnglePreserving();
	VM3Point p(0,0,0);
   VM3Point q = iM * p;
	return (q);

#if 0
	//////////////
	// this is the old round-about way of computing the same as above...
	//////////////

   // extract the after rotation translation component
   VM3Point p(mat[0][3], mat[1][3], mat[2][3]);

   // extract the rotation component and invert it (use invertAnglePreserving
   // because it is much faster and since a simple R + T matrix, can use it)
   VM4Matrix iR = *this;
   iR.mat[0][3] = 0.0;
   iR.mat[1][3] = 0.0;
   iR.mat[2][3] = 0.0;
   iR.invertAnglePreserving();

   // apply the inverted rotation transformation the extracted translation
   // component; this will compute the equivalent translation in the pre-
   // rotated world. the -1 is to make it go from origin to the camera
   VM3Point q = -1 * (iR * p);

   return (q);
#endif

}


/*****************************************************************************

  Very simple method, just apply the mat (and its rotations) to the original
  directional vector. The -1 is rotate from origDir to dir.

  This method is useful, for instance, to extract the camera direction from a
  standard OpenGL model transformation matrix (note: this code, as all vmmatrix
  code, assumes column-major vectors so it is really the transpose of the
  OpenGL matrices).

*****************************************************************************/
VM3Vector
VM4Matrix::cameraDirection(VM3Vector origDir)
{
   // make sure bottom row is 0,0,0,1
   if (mat[3][0] != 0 || mat[3][1] != 0 || mat[3][2] != 0 || mat[3][3] != 1) {
      return VM3Vector(0,0,0);
   }

	// by default OpenGL goes from world space to canonical space, so to find 
	// the direction of the camera vector origDir in world space must invert 
	// the matrix
	VM4Matrix iM = *this;
	iM.invertAnglePreserving();
   VM3Vector dir = iM * origDir;
   return (dir);

#if 0
	//////////////
	// this is the old round-about way of computing the same as above...
	//////////////

   VM4Matrix iR = *this;
	iR.mat[0][3] = 0.0;
	iR.mat[1][3] = 0.0;
	iR.mat[2][3] = 0.0;
	iR.invertAnglePreserving();
   VM3Vector dir = iR * origDir;
   return (dir);
#endif
}


/*****************************************************************************

* Apply x first, then y, then z.

*****************************************************************************/
void
VM4Matrix::zyxRotationTheta(double *z, double *y, double *x)
{
	double thetaX, thetaY, thetaZ;

	thetaY = asin(-mat[2][0]);
	if (thetaY < M_PI/2) {
		if (thetaY > -M_PI/2) {
			thetaZ = atan2(mat[1][0], mat[0][0]);
			thetaX = atan2(mat[2][1], mat[2][2]);
		} else {
			// not a unique solution
			thetaZ = -atan2(mat[0][1], mat[0][2]);
			thetaX = 0;
		}
	} else {
		// not a unique solution
		thetaZ = atan2(-mat[0][1], mat[0][2]);
		thetaX = 0;
	}

	*z = thetaZ;
	*y = thetaY;
	*x = thetaX;
}
/****************************************************************************
* Apply z first, then y, then x.
*****************************************************************************/
void
VM4Matrix::xyzRotationTheta(double *x, double *y, double *z)
{
	double thetaX, thetaY, thetaZ;

	thetaY = asin(mat[0][2]);
	if (thetaY < M_PI/2) {
		if (thetaY > -M_PI/2) {
			thetaX = atan2(-mat[1][2], mat[2][2]);
			thetaZ = atan2(-mat[0][1], mat[0][0]);
		} else {
			// not a unique solution
			thetaX = -atan2(mat[1][0], mat[1][1]);
			thetaZ = 0;
		}
	} else {
		// not a unique solution
		thetaX = atan2(mat[1][0], mat[1][1]);
		thetaZ = 0;
	}

	*z = thetaZ;
	*y = thetaY;
	*x = thetaX;
}

