#ifndef VM4MATRIX_ALREADY_INCLUDED
#define VM4MATRIX_ALREADY_INCLUDED

#include <vm3vector.h>
#include <vm3point.h>


#define VM4MATRIX_ROTATE_X_AXIS        0
#define VM4MATRIX_ROTATE_Y_AXIS        1
#define VM4MATRIX_ROTATE_Z_AXIS        2

#define VM4MATRIX_RAD_TO_DEG(R)        (((R)/M_PI)*180.0)
#define VM4MATRIX_DEG_TO_RAD(D)        (((D)/180.0)*M_PI)


/*****************************************************************************

This class implements a standard 4x4 matrix. Operations are assumed to be 
performed with homogenous 3D coordinates (using VM3Vector/VM3Point). Supported
operations are: translate, scale, rotate, invert.

*****************************************************************************/

class VM4Matrix {
   private:
      float mat[4][4];

   public:

      /* 
       * Constructors 
       */
      VM4Matrix() 
      { 
         mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0; mat[0][3] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0; mat[1][3] = 0.0;
         mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0; mat[2][3] = 0.0;
         mat[3][0] = 0.0; mat[3][1] = 0.0; mat[3][2] = 0.0; mat[3][3] = 1.0;
      }

      VM4Matrix(float a, float b, float c, float d, 
                float e, float f, float g, float h, 
                float i, float j, float k, float l,
                float m, float n, float o, float p) 
      {
         mat[0][0] = a; mat[0][1] = b; mat[0][2] = c; mat[0][3] = d;
         mat[1][0] = e; mat[1][1] = f; mat[1][2] = g; mat[1][3] = h;
         mat[2][0] = i; mat[2][1] = j; mat[2][2] = k; mat[2][3] = l;
         mat[3][0] = m; mat[3][1] = n; mat[3][2] = o; mat[3][3] = p;
      }

      VM4Matrix(float m[4][4]) 
      {
         BCOPY(m, mat, sizeof(float)*16);
      }

      VM4Matrix(const VM4Matrix &m) 
      { 
         BCOPY(m.mat, mat, sizeof(float)*16);
      }


      /*
       * Matrix operations and utilities
       */
      void identity() 
      {
         mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0; mat[0][3] = 0.0;
         mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0; mat[1][3] = 0.0;
         mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0; mat[2][3] = 0.0;
         mat[3][0] = 0.0; mat[3][1] = 0.0; mat[3][2] = 0.0; mat[3][3] = 1.0;
      }

      void zero() 
      {
         BZERO(mat, sizeof(float)*16);
      }

      void transpose() 
      {
        int   r,c;
        float tmp;

        for (r = 0; r < 4; r++) {
           for (c = r; c < 4; c++) {
              if (r != c) {
                 SWAP(mat[r][c], mat[c][r], tmp);
              }
           }
        }
      }

      /* translate operation */
      void translate(float x, float y, float z) 
      {
         mat[0][3] += x;
         mat[1][3] += y;
         mat[2][3] += z;
      }
      void translate(VM3Point  &t) { translate(t.x(),t.y(),t.z()); }
      void translate(VM3Vector &t) { translate(t.x(),t.y(),t.z()); }

      /* scale operation */
      void scale(float x, float y, float z) 
      {
         VM4Matrix mat;

         mat.mat[0][0] = x;
         mat.mat[1][1] = y;
         mat.mat[2][2] = z;

         *this = (*this) * mat;
      }
      void scale(const VM3Point  &s) { scale(s.x(),s.y(),s.z()); }
      void scale(const VM3Vector &s) { scale(s.x(),s.y(),s.z()); }
      void scale(float s)      { scale(s, s, s); }

      /* rotation operation */
      void rotate(int axis, float radians);
      void rotate(const VM3Vector &axis, float radians);
      void rotate(float x, float y, float z, float radians)
      { 
         VM3Vector axis(x,y,z);
         rotate(axis, radians); 
      }

      /* inversion operation */
      int  invert();                      // invert self
      VM4Matrix inverse()                 // return inverse as a matrix
		{
			VM4Matrix tmp = *this;
			tmp.invert();
			return (tmp);
		}
      int  invertAnglePreserving();       // invert self
      VM4Matrix inverseAnglePreserving()  // return inverse as a matrix
		{
			VM4Matrix tmp = *this;
			tmp.invertAnglePreserving();
			return (tmp);
		}

      /* get camera position from a (graphics) transformation matrix */
      VM3Point cameraPosition();

      /* get camera direction from a (graphics) transformation matrix; the
         origDir tells the method how the camera was oriented when at the 
         origin, e.g. (0,0,-1) */
      VM3Vector cameraDirection(VM3Vector origDir);

		/* get the z,y,x rotation radians from the rotation component */
      void zyxRotationTheta(double *z, double *y, double *x);
      void xyzRotationTheta(double *x, double *y, double *z);
      
      /* others */
      void set(int r, int c, float val) 
      { 
         IFDEBUG(r < 0 || r >= 4 || c < 0 || c >= 4) {
            fprintf(stderr, "VM4Matrix: index (%i,%i) out of range\n", r, c);
            return;
         }
         mat[r][c] = val; 
      }
      void print(FILE *fp)
      {
         int i,j;

         for (i=0; i<4; i++) {
            for (j=0; j<4; j++)
               fprintf(fp,"%1.4f ", mat[i][j]);
            fprintf(fp,"\n");
         }
         fprintf(fp,"\n");
      }
      float *data() { return (float *)mat; }



      /*
       * Operator overloading
       */
      inline float operator()(int r, int c) const
      {
         IFDEBUG(r < 0 || r >= 4 || c < 0 || c >= 4) {
            fprintf(stderr, "VM4Matrix: index (%i,%i) out of range\n", r, c);
            return 0;
         }
         return (mat[r][c]);
      }

      inline VM4Matrix &operator=(const VM4Matrix &m) 
      {
         BCOPY(m.mat, mat, sizeof(float)*16);
         return *this;
      }

      inline VM4Matrix &operator+=(const VM4Matrix &m) 
      {
         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             mat[i][j] += m.mat[i][j];

         return *this;
      }

      inline VM4Matrix &operator-=(const VM4Matrix &m) 
      {
         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             mat[i][j] -= m.mat[i][j];

         return *this;
      }

      inline VM4Matrix &operator*=(const VM4Matrix &m) 
      {
         register float res[4][4];

         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             res[i][j] = mat[i][0] * m.mat[0][j] +
                         mat[i][1] * m.mat[1][j] +
                         mat[i][2] * m.mat[2][j] +
                         mat[i][3] * m.mat[3][j];
         BCOPY(res, mat, sizeof(float)*16);
         return *this;
      }

      inline friend VM4Matrix operator+(const VM4Matrix &a, const VM4Matrix &b) 
      {
         register float res[4][4];

         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             res[i][j] = a.mat[i][j] + b.mat[i][j];
         return (VM4Matrix(res));
      }

      inline friend VM4Matrix operator-(const VM4Matrix &a, const VM4Matrix &b) 
      {
         register float res[4][4];

         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             res[i][j] = a.mat[i][j] - b.mat[i][j];
         return (VM4Matrix(res));
      }

      inline friend VM4Matrix operator*(const VM4Matrix &a, const VM4Matrix &b)
      {
         register float res[4][4];

         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             res[i][j] = a.mat[i][0] * b.mat[0][j] +
                         a.mat[i][1] * b.mat[1][j] +
                         a.mat[i][2] * b.mat[2][j] +
                         a.mat[i][3] * b.mat[3][j];
         return (VM4Matrix(res));
      }

      inline friend VM4Matrix operator*(const VM4Matrix &a, float b)
      {
         register float res[4][4];

         for (int i=0; i<4; i++)
          for (int j=0; j<4; j++)
             res[i][j] = a.mat[i][j] * b;
         return VM4Matrix(res);
      }

      inline int operator==(const VM4Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; ret && i<4; i++)
            for (int j=0; ret && j<4; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = FALSE;
         return (ret);
      }

      inline int operator!=(const VM4Matrix &a) 
      {
         register int ret=TRUE;

         for (int i=0; !ret && i<4; i++)
            for (int j=0; !ret && j<4; j++)
               if (ABS(mat[i][j] - a.mat[i][j]) > FLOAT_EPSILON)
                  ret = TRUE;
         return (ret);
      }

      /*
       * Destructors
       */
      ~VM4Matrix() {};
};


#endif
