#ifndef VM2POINT_ALREADY_INCLUDED
#define VM2POINT_ALREADY_INCLUDED

#include <common.h>

#define VM2POINT_X       0
#define VM2POINT_Y       1

class VM2Vector;
class VM2Matrix;
class VM3Matrix;


/*****************************************************************************

This object stores the information that defines a point in 2 space.

*****************************************************************************/

class VM2Point {
   private:
      float px, py;

   public:
      /*
       * constructors are overloaded
       */
      VM2Point() 
      {
         px = py = 0.0;
      }

      VM2Point(VM2Vector &);

      VM2Point(const VM2Point &ref) 
      { 
         px = ref.px; py = ref.py;
      }

      VM2Point(int x, int y)
      { 
         px = (float)x; py = (float)y;
      }

      VM2Point(float x, float y)
      { 
         px = x; py = y;
      }

      VM2Point(double x, double y)
      { 
         px = x; py = y;
      }

      VM2Point(float p[2]) 
      {
         px = p[0]; py = p[1];
      }


      /*
       * utilities
       */ 
      inline void set(float sx, float sy)
      {
         px = sx; py = sy;
      }

      inline void negate() 
      {
         px = -px; py = -py; 
      }

      inline void print(FILE *fp) 
      {
         fprintf(fp,"(%1.4f, %1.4f)\n",px,py);
      }

      inline float x() const { return px; }; 
      inline float y() const { return py; }; 

      inline int isZero() 
      { 
         return ((ABS(px - 0) < FLOAT_EPSILON) &&
                 (ABS(py - 0) < FLOAT_EPSILON));
      }


      /*
       * operator overloading
       */
      inline float operator()(int r) const
      { 
         float val;
         switch (r) {
            case VM2POINT_X : val = px; break;
            case VM2POINT_Y : val = py; break;
            default: fprintf(stderr, 
                             "VM2Point: coordinate index out of range\n");
                     val = 0;
         }
         return (val);
      }

      inline VM2Point &operator=(const VM2Point &c) 
      {
         px = c.px; py = c.py; 
         return *this;
      }

      inline VM2Point &operator+=(const VM2Point& a) 
      {
         px += a.px; py += a.py;
         return *this;
      }

      inline VM2Point &operator-=(const VM2Point& a) 
      {
         px -= a.px; py -= a.py;
         return *this;
      }

      inline friend VM2Point operator+(const VM2Point& a, const VM2Point& b) 
      {
         return VM2Point(a.px+b.px, a.py+b.py);
      }

      inline friend VM2Point operator+(const VM2Point& a, const float v) 
      {
         return VM2Point(a.px+v, a.py+v);
      }

      inline friend VM2Point operator-(const VM2Point& a, const VM2Point& b) 
      {
         return VM2Point(a.px-b.px, a.py-b.py);
      } 

      inline friend VM2Point operator-(const VM2Point& a, const float v) 
      {
         return VM2Point(a.px-v, a.py-v);
      } 

      inline friend VM2Point operator*(const VM2Point& a, int s) 
      {
         return VM2Point(a.px*(float)s, a.py*(float)s);
      }

      inline friend VM2Point operator*(const VM2Point& a, float s) 
      {
         return VM2Point(a.px*s, a.py*s);
      }

      inline friend VM2Point operator*(const VM2Point& a, double s) 
      {
         return VM2Point((float)(a.px*s), (float)(a.py*s));
      }

      inline friend VM2Point operator*(int s, const VM2Point& a) 
      {
         return VM2Point(a.px*(float)s, a.py*(float)s);
      }

      inline friend VM2Point operator*(float s, const VM2Point& a) 
      {
         return VM2Point(a.px*s, a.py*s);
      }

      inline friend VM2Point operator*(double s, const VM2Point& a) 
      {
         return VM2Point((float)(a.px*s), (float)(a.py*s));
      }

      inline friend VM2Point operator/(const VM2Point& a, int s) 
      {
         return VM2Point(a.px/(float)s, a.py/(float)s);
      }

      inline friend VM2Point operator/(const VM2Point& a, float s) 
      {
         return VM2Point(a.px/s, a.py/s);
      }

      inline friend VM2Point operator/(const VM2Point& a, double s) 
      {
         return VM2Point(a.px/s, a.py/s);
      }


      /************/

      friend VM2Point operator*(const VM2Matrix& m, const VM2Point &p);
      friend VM2Point operator*(const VM3Matrix& m, const VM2Point &p);

      /************/

      inline int operator==(const VM2Point& p) {
         return (ABS(px - p.px) < FLOAT_EPSILON && 
                 ABS(py - p.py) < FLOAT_EPSILON);
      }

      inline int operator!=(const VM2Point& p) {
         return (ABS(px - p.px) > FLOAT_EPSILON ||
                 ABS(py - p.py) > FLOAT_EPSILON);
      }

      inline int operator>=(const VM2Point& v) 
      {
         return ((px >= v.px) && (py >= v.py)) ? TRUE : FALSE;
      }

      inline int operator<=(const VM2Point& v) 
      {
         return ((px <= v.px) && (py <= v.py)) ? TRUE : FALSE;
      }

      inline int operator<(const VM2Point& v) 
      {
         return ((px < v.px) && (py < v.py)) ? TRUE : FALSE;
      }

      inline int operator>(const VM2Point& v) 
      {
         return ((px > v.px) && (py > v.py)) ? TRUE : FALSE;
      }
};

#endif
